# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0.

from awscrt.io import ClientBootstrap, ClientTlsContext, DefaultHostResolver, EventLoopGroup, Pkcs11Lib, TlsContextOptions
from awscrt import http
from awscrt.mqtt import Client, Connection, QoS, Will, OnConnectionClosedData, OnConnectionFailureData, OnConnectionSuccessData, ConnectReturnCode
from test import test_retry_wrapper, NativeResourceTest
from concurrent.futures import Future
import os
import unittest
import uuid
import time

TIMEOUT = 100.0


def _get_env_variable(env_name):
    env_data = os.environ.get(env_name)
    if not env_data:
        raise unittest.SkipTest(f"test requires env var: {env_name}")
    return env_data


def create_client_id():
    return 'aws-crt-python-unit-test-{0}'.format(uuid.uuid4())


class MqttConnectionTest(NativeResourceTest):
    TEST_TOPIC = '/test/me/senpai/' + str(uuid.uuid4())
    TEST_MSG = 'NOTICE ME!'.encode('utf8')

    def _create_connection(
            self,
            endpoint,
            tls_context,
            port=8883,
            use_static_singletons=False,
            client_id=None,
            on_connection_success_callback=None,
            on_connection_failure_callback=None,
            on_connection_closed_callback=None,
            on_connection_resumed_callback=None):
        if use_static_singletons:
            client = Client(tls_ctx=tls_context)
        else:
            elg = EventLoopGroup()
            resolver = DefaultHostResolver(elg)
            bootstrap = ClientBootstrap(elg, resolver)
            client = Client(bootstrap, tls_context)

        connection = Connection(
            client=client,
            client_id=client_id if client_id else create_client_id(),
            host_name=endpoint,
            port=port,
            on_connection_closed=on_connection_closed_callback,
            on_connection_failure=on_connection_failure_callback,
            on_connection_success=on_connection_success_callback,
            on_connection_resumed=on_connection_resumed_callback)
        return connection

    def _test_connect_disconnect(self):
        test_input_endpoint = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_HOST")
        test_input_cert = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_CERT")
        test_input_key = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_KEY")

        test_tls_opts = TlsContextOptions.create_client_with_mtls_from_path(test_input_cert, test_input_key)
        test_tls = ClientTlsContext(test_tls_opts)

        connection = self._create_connection(test_input_endpoint, test_tls)
        connection.connect().result(TIMEOUT)
        connection.disconnect().result(TIMEOUT)

    def test_connect_disconnect(self):
        test_retry_wrapper(self._test_connect_disconnect)

    def _test_ecc_connect_disconnect(self):
        test_input_endpoint = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_HOST")
        test_input_cert = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_ECC_CERT")
        test_input_key = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_ECC_KEY")

        test_tls_opts = TlsContextOptions.create_client_with_mtls_from_path(test_input_cert, test_input_key)
        test_tls = ClientTlsContext(test_tls_opts)

        connection = self._create_connection(test_input_endpoint, test_tls)
        connection.connect().result(TIMEOUT)
        connection.disconnect().result(TIMEOUT)

    def test_ecc_connect_disconnect(self):
        test_retry_wrapper(self._test_ecc_connect_disconnect)

    def _test_pkcs11(self):
        test_input_endpoint = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_HOST")
        test_input_pkcs11_lib = _get_env_variable("AWS_TEST_PKCS11_LIB")
        test_input_pkcs11_pin = _get_env_variable("AWS_TEST_PKCS11_PIN")
        test_input_pkcs11_token_label = _get_env_variable("AWS_TEST_PKCS11_TOKEN_LABEL")
        test_input_pkcs11_private_key = _get_env_variable("AWS_TEST_PKCS11_PKEY_LABEL")
        test_input_pkcs11_cert = _get_env_variable("AWS_TEST_PKCS11_CERT_FILE")

        test_pkcs11_lib = Pkcs11Lib(file=test_input_pkcs11_lib, behavior=Pkcs11Lib.InitializeFinalizeBehavior.STRICT)
        test_tls_opts = TlsContextOptions.create_client_with_mtls_pkcs11(
            pkcs11_lib=test_pkcs11_lib,
            user_pin=test_input_pkcs11_pin,
            token_label=test_input_pkcs11_token_label,
            private_key_label=test_input_pkcs11_private_key,
            cert_file_path=test_input_pkcs11_cert
        )
        test_tls = ClientTlsContext(test_tls_opts)

        connection = self._create_connection(test_input_endpoint, test_tls)
        connection.connect().result(TIMEOUT)
        connection.disconnect().result(TIMEOUT)

    def test_pkcs11(self):
        test_retry_wrapper(self._test_pkcs11)

    def _test_pub_sub(self):
        self.TEST_TOPIC = '/test/me/senpai/' + str(uuid.uuid4())
        test_input_endpoint = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_HOST")
        test_input_cert = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_CERT")
        test_input_key = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_KEY")
        test_tls_opts = TlsContextOptions.create_client_with_mtls_from_path(test_input_cert, test_input_key)
        test_tls = ClientTlsContext(test_tls_opts)
        connection = self._create_connection(test_input_endpoint, test_tls)

        connection.connect().result(TIMEOUT)
        received = Future()

        def on_message(**kwargs):
            received.set_result(kwargs)

        # subscribe
        subscribed, packet_id = connection.subscribe(self.TEST_TOPIC, QoS.AT_LEAST_ONCE, on_message)
        suback = subscribed.result(TIMEOUT)
        self.assertEqual(packet_id, suback['packet_id'])
        self.assertEqual(self.TEST_TOPIC, suback['topic'])
        self.assertIs(QoS.AT_LEAST_ONCE, suback['qos'])

        # publish
        published, packet_id = connection.publish(self.TEST_TOPIC, self.TEST_MSG, QoS.AT_LEAST_ONCE)
        puback = published.result(TIMEOUT)
        self.assertEqual(packet_id, puback['packet_id'])

        # receive message
        rcv = received.result(TIMEOUT)
        self.assertEqual(self.TEST_TOPIC, rcv['topic'])
        self.assertEqual(self.TEST_MSG, rcv['payload'])
        self.assertFalse(rcv['dup'])
        self.assertEqual(QoS.AT_LEAST_ONCE, rcv['qos'])
        self.assertFalse(rcv['retain'])

        # unsubscribe
        unsubscribed, packet_id = connection.unsubscribe(self.TEST_TOPIC)
        unsuback = unsubscribed.result(TIMEOUT)
        self.assertEqual(packet_id, unsuback['packet_id'])

        # disconnect
        connection.disconnect().result(TIMEOUT)

    def test_pub_sub(self):
        test_retry_wrapper(self._test_pub_sub)

    def _test_will(self):
        self.TEST_TOPIC = '/test/me/senpai/' + str(uuid.uuid4())
        test_input_endpoint = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_HOST")
        test_input_cert = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_CERT")
        test_input_key = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_KEY")
        test_tls_opts = TlsContextOptions.create_client_with_mtls_from_path(test_input_cert, test_input_key)
        test_tls = ClientTlsContext(test_tls_opts)

        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)
        client = Client(bootstrap, test_tls)

        will_client_id = create_client_id()
        connection = Connection(
            client=client,
            client_id=will_client_id,
            host_name=test_input_endpoint,
            port=8883,
            will=Will(self.TEST_TOPIC, QoS.AT_LEAST_ONCE, self.TEST_MSG, False),
            ping_timeout_ms=10000,
            keep_alive_secs=30
        )
        connection.connect().result(TIMEOUT)

        subscriber = Connection(
            client=client,
            client_id=create_client_id(),
            host_name=test_input_endpoint,
            port=8883,
            ping_timeout_ms=10000,
            keep_alive_secs=30
        )
        subscriber.connect().result(TIMEOUT)

        received = Future()

        def on_message(**kwargs):
            received.set_result(kwargs)

        # subscribe
        subscribed, packet_id = subscriber.subscribe(self.TEST_TOPIC, QoS.AT_LEAST_ONCE, on_message)
        suback = subscribed.result(TIMEOUT)
        self.assertEqual(packet_id, suback['packet_id'])
        self.assertEqual(self.TEST_TOPIC, suback['topic'])
        self.assertIs(QoS.AT_LEAST_ONCE, suback['qos'])

        # wait a few seconds to ensure we can make another client ID connect (don't trigger IoT Core limit)
        time.sleep(2)

        # Disconnect the will client to send the will by making another connection with the same client id
        disconnecter = Connection(
            client=client,
            client_id=will_client_id,
            host_name=test_input_endpoint,
            port=8883,
            will=Will(self.TEST_TOPIC, QoS.AT_LEAST_ONCE, self.TEST_MSG, False),
            ping_timeout_ms=10000,
            keep_alive_secs=30
        )

        # A race condition exists in IoT Core where the interrupter may get refused rather than the existing
        # connection getting dropped.  Loop until we successfully connect.
        continue_connecting = True
        while continue_connecting:
            try:
                disconnecter.connect().result(TIMEOUT)
                continue_connecting = False
            except BaseException:
                pass

        # Receive message
        rcv = received.result(TIMEOUT)
        self.assertEqual(self.TEST_TOPIC, rcv['topic'])
        self.assertEqual(self.TEST_MSG, rcv['payload'])
        self.assertFalse(rcv['dup'])
        self.assertEqual(QoS.AT_LEAST_ONCE, rcv['qos'])
        self.assertFalse(rcv['retain'])

        # Disconnect the other clients
        connection.disconnect().result(TIMEOUT)
        disconnecter.disconnect().result(TIMEOUT)

        # unsubscribe
        unsubscribed, packet_id = subscriber.unsubscribe(self.TEST_TOPIC)
        unsuback = unsubscribed.result(TIMEOUT)
        self.assertEqual(packet_id, unsuback['packet_id'])

        # disconnect
        subscriber.disconnect().result(TIMEOUT)

    def test_will(self):
        test_retry_wrapper(self._test_will)

    def _test_on_message(self):
        test_input_endpoint = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_HOST")
        test_input_cert = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_CERT")
        test_input_key = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_KEY")
        test_tls_opts = TlsContextOptions.create_client_with_mtls_from_path(test_input_cert, test_input_key)
        test_tls = ClientTlsContext(test_tls_opts)
        connection = self._create_connection(test_input_endpoint, test_tls)

        received = Future()

        def on_message(**kwargs):
            received.set_result(kwargs)

        # on_message for connection has to be set before connect, or possible race will happen
        connection.on_message(on_message)

        connection.connect().result(TIMEOUT)
        # subscribe without callback
        subscribed, packet_id = connection.subscribe(self.TEST_TOPIC, QoS.AT_LEAST_ONCE)
        subscribed.result(TIMEOUT)

        # publish
        published, packet_id = connection.publish(self.TEST_TOPIC, self.TEST_MSG, QoS.AT_LEAST_ONCE)
        puback = published.result(TIMEOUT)

        # receive message
        rcv = received.result(TIMEOUT)
        self.assertEqual(self.TEST_TOPIC, rcv['topic'])
        self.assertEqual(self.TEST_MSG, rcv['payload'])
        self.assertFalse(rcv['dup'])
        self.assertEqual(QoS.AT_LEAST_ONCE, rcv['qos'])
        self.assertFalse(rcv['retain'])

        # disconnect
        connection.disconnect().result(TIMEOUT)

    def test_on_message(self):
        test_retry_wrapper(self._test_on_message)

    def _test_on_message_old_fn_signature(self):
        # ensure that message-received callbacks with the old function signature still work

        test_input_endpoint = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_HOST")
        test_input_cert = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_CERT")
        test_input_key = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_KEY")
        test_tls_opts = TlsContextOptions.create_client_with_mtls_from_path(test_input_cert, test_input_key)
        test_tls = ClientTlsContext(test_tls_opts)
        connection = self._create_connection(test_input_endpoint, test_tls)

        any_received = Future()
        sub_received = Future()

        # Note: Testing degenerate callback signature that failed to take
        # forward-compatibility **kwargs.
        def on_any_message(topic, payload):
            any_received.set_result({'topic': topic, 'payload': payload})

        def on_sub_message(topic, payload):
            sub_received.set_result({'topic': topic, 'payload': payload})

        # on_message for connection has to be set before connect, or possible race will happen
        connection.on_message(on_any_message)

        connection.connect().result(TIMEOUT)
        # subscribe without callback
        subscribed, packet_id = connection.subscribe(self.TEST_TOPIC, QoS.AT_LEAST_ONCE, on_sub_message)
        subscribed.result(TIMEOUT)

        # publish
        published, packet_id = connection.publish(self.TEST_TOPIC, self.TEST_MSG, QoS.AT_LEAST_ONCE)
        puback = published.result(TIMEOUT)

        # receive message
        rcv = any_received.result(TIMEOUT)
        self.assertEqual(self.TEST_TOPIC, rcv['topic'])
        self.assertEqual(self.TEST_MSG, rcv['payload'])

        rcv = sub_received.result(TIMEOUT)
        self.assertEqual(self.TEST_TOPIC, rcv['topic'])
        self.assertEqual(self.TEST_MSG, rcv['payload'])

        # disconnect
        connection.disconnect().result(TIMEOUT)

    def test_on_message_old_fn_signature(self):
        test_retry_wrapper(self._test_on_message_old_fn_signature)

    def _test_connect_disconnect_with_default_singletons(self):
        test_input_endpoint = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_HOST")
        test_input_cert = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_CERT")
        test_input_key = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_KEY")

        test_tls_opts = TlsContextOptions.create_client_with_mtls_from_path(test_input_cert, test_input_key)
        test_tls = ClientTlsContext(test_tls_opts)

        connection = self._create_connection(test_input_endpoint, test_tls, use_static_singletons=True)
        connection.connect().result(TIMEOUT)
        connection.disconnect().result(TIMEOUT)

        # free singletons
        ClientBootstrap.release_static_default()
        EventLoopGroup.release_static_default()
        DefaultHostResolver.release_static_default()

    def test_connect_disconnect_with_default_singletons(self):
        test_retry_wrapper(self._test_connect_disconnect_with_default_singletons)

    def _test_connect_publish_wait_statistics_disconnect(self):
        test_input_endpoint = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_HOST")
        test_input_cert = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_CERT")
        test_input_key = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_KEY")
        test_tls_opts = TlsContextOptions.create_client_with_mtls_from_path(test_input_cert, test_input_key)
        test_tls = ClientTlsContext(test_tls_opts)
        connection = self._create_connection(test_input_endpoint, test_tls)

        connection.connect().result(TIMEOUT)

        # check operation statistics
        statistics = connection.get_stats()
        self.assertEqual(statistics.incomplete_operation_count, 0)
        self.assertEqual(statistics.incomplete_operation_size, 0)
        self.assertEqual(statistics.unacked_operation_count, 0)
        self.assertEqual(statistics.unacked_operation_size, 0)

        # publish
        published, packet_id = connection.publish(self.TEST_TOPIC, self.TEST_MSG, QoS.AT_LEAST_ONCE)
        puback = published.result(TIMEOUT)
        self.assertEqual(packet_id, puback['packet_id'])

        # check operation statistics
        statistics = connection.get_stats()
        self.assertEqual(statistics.incomplete_operation_count, 0)
        self.assertEqual(statistics.incomplete_operation_size, 0)
        self.assertEqual(statistics.unacked_operation_count, 0)
        self.assertEqual(statistics.unacked_operation_size, 0)

        # disconnect
        connection.disconnect().result(TIMEOUT)

    def test_connect_publish_wait_statistics_disconnect(self):
        test_retry_wrapper(self._test_connect_publish_wait_statistics_disconnect)

    def _test_connect_publish_statistics_wait_disconnect(self):
        test_input_endpoint = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_HOST")
        test_input_cert = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_CERT")
        test_input_key = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_KEY")
        test_tls_opts = TlsContextOptions.create_client_with_mtls_from_path(test_input_cert, test_input_key)
        test_tls = ClientTlsContext(test_tls_opts)
        connection = self._create_connection(test_input_endpoint, test_tls)

        connection.connect().result(TIMEOUT)

        # publish
        published, packet_id = connection.publish(self.TEST_TOPIC, self.TEST_MSG, QoS.AT_LEAST_ONCE)
        # Per packet: (The size of the topic, the size of the payload, 2 for the header and 2 for the packet ID)
        expected_size = len(self.TEST_TOPIC) + len(self.TEST_MSG) + 4

        # check operation statistics
        statistics = connection.get_stats()
        self.assertEqual(statistics.incomplete_operation_count, 1)
        self.assertEqual(statistics.incomplete_operation_size, expected_size)
        # NOTE: Unacked MAY be zero because we have not invoked the future yet
        # and so it has not had time to move to the socket. With Python especially, it seems to heavily depend on how fast
        # the test is executed, which makes it sometimes rarely get into the unacked operation and then it is non-zero.
        # To fix this, we just make sure it is within expected bounds (0 or within packet size).
        self.assertTrue(statistics.unacked_operation_count <= 1)
        self.assertTrue(statistics.unacked_operation_count <= expected_size)

        # wait for PubAck
        puback = published.result(TIMEOUT)
        self.assertEqual(packet_id, puback['packet_id'])

        # check operation statistics
        statistics = connection.get_stats()
        self.assertEqual(statistics.incomplete_operation_count, 0)
        self.assertEqual(statistics.incomplete_operation_size, 0)
        self.assertEqual(statistics.unacked_operation_count, 0)
        self.assertEqual(statistics.unacked_operation_size, 0)

        # disconnect
        connection.disconnect().result(TIMEOUT)

    def test_connect_publish_statistics_wait_disconnect(self):
        test_retry_wrapper(self._test_connect_publish_statistics_wait_disconnect)

    def _test_connect_disconnect_with_callbacks_happy(self):
        test_input_endpoint = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_HOST")
        test_input_cert = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_CERT")
        test_input_key = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_KEY")
        test_tls_opts = TlsContextOptions.create_client_with_mtls_from_path(test_input_cert, test_input_key)
        test_tls = ClientTlsContext(test_tls_opts)

        on_connection_success_future = Future()
        on_connection_closed_future = Future()

        def on_connection_success_callback(connection, callback_data: OnConnectionSuccessData):
            on_connection_success_future.set_result(
                {'return_code': callback_data.return_code, "session_present": callback_data.session_present})

        def on_connection_failure_callback(connection, callback_data: OnConnectionFailureData):
            pass

        def on_connection_closed_callback(connection, callback_data: OnConnectionClosedData):
            on_connection_closed_future.set_result({})

        connection = self._create_connection(
            endpoint=test_input_endpoint,
            tls_context=test_tls,
            on_connection_success_callback=on_connection_success_callback,
            on_connection_failure_callback=on_connection_failure_callback,
            on_connection_closed_callback=on_connection_closed_callback)
        connection.connect().result(TIMEOUT)
        success_data = on_connection_success_future.result(TIMEOUT)
        self.assertEqual(success_data['return_code'], ConnectReturnCode.ACCEPTED)
        self.assertEqual(success_data['session_present'], False)
        connection.disconnect().result(TIMEOUT)
        on_connection_closed_future.result(TIMEOUT)

    def test_connect_disconnect_with_callbacks_happy(self):
        test_retry_wrapper(self._test_connect_disconnect_with_callbacks_happy)

    def _test_connect_disconnect_with_callbacks_unhappy(self):
        test_input_endpoint = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_HOST")
        test_input_cert = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_CERT")
        test_input_key = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_KEY")
        test_tls_opts = TlsContextOptions.create_client_with_mtls_from_path(test_input_cert, test_input_key)
        test_tls = ClientTlsContext(test_tls_opts)

        on_onnection_failure_future = Future()

        def on_connection_success_callback(connection, callback_data: OnConnectionSuccessData):
            pass

        def on_connection_failure_callback(connection, callback_data: OnConnectionFailureData):
            on_onnection_failure_future.set_result({'error': callback_data.error})

        def on_connection_closed_callback(connection, callback_data: OnConnectionClosedData):
            pass

        connection = self._create_connection(
            endpoint=test_input_endpoint,
            tls_context=test_tls,
            port=1234,
            on_connection_success_callback=on_connection_success_callback,
            on_connection_failure_callback=on_connection_failure_callback,
            on_connection_closed_callback=on_connection_closed_callback)

        exception_occurred = False
        try:
            connection.connect().result(TIMEOUT)
        except Exception:
            exception_occurred = True
        self.assertTrue(exception_occurred, "Exception did not occur when connecting with invalid arguments!")

        failure_data = on_onnection_failure_future.result(TIMEOUT)
        self.assertTrue(failure_data['error'] is not None)

    def test_connect_disconnect_with_callbacks_unhappy(self):
        test_retry_wrapper(self._test_connect_disconnect_with_callbacks_unhappy)

    def _test_connect_disconnect_with_callbacks_happy_on_resume(self):
        # Check that an on_connection_success callback fires on a resumed connection.

        # NOTE Since there is no mocked server available on this abstraction level, the only sensible approach
        # is to interrupt a connection, and wait for it to be resumed automatically. For that, another client
        # with the same client_id connects to the server and then immediately disconnects.

        test_input_endpoint = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_HOST")
        test_input_cert = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_CERT")
        test_input_key = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_KEY")
        test_tls_opts = TlsContextOptions.create_client_with_mtls_from_path(test_input_cert, test_input_key)
        test_tls = ClientTlsContext(test_tls_opts)

        on_connection_success_future = Future()
        on_connection_closed_future = Future()
        on_connection_resumed_future = Future()

        def on_connection_success_callback(connection, callback_data: OnConnectionSuccessData):
            on_connection_success_future.set_result(
                {'return_code': callback_data.return_code, "session_present": callback_data.session_present})

        def on_connection_closed_callback(connection, callback_data: OnConnectionClosedData):
            on_connection_closed_future.set_result({})

        def on_connection_resumed_callback(connection, return_code: ConnectReturnCode, session_present):
            on_connection_resumed_future.set_result(
                {'return_code': return_code, "session_present": session_present})

        connection = self._create_connection(
            endpoint=test_input_endpoint,
            tls_context=test_tls,
            on_connection_success_callback=on_connection_success_callback,
            on_connection_closed_callback=on_connection_closed_callback,
            on_connection_resumed_callback=on_connection_resumed_callback)
        connection.connect().result(TIMEOUT)
        success_data = on_connection_success_future.result(TIMEOUT)
        self.assertEqual(success_data['return_code'], ConnectReturnCode.ACCEPTED)
        self.assertEqual(success_data['session_present'], False)

        # Putting a sleep here helps prevent a "race" condition in IoT Core where the second connection can get
        # rejected rather than the first disconnected.
        time.sleep(5)

        # Reset the future for the reconnect attempt.
        on_connection_success_future = Future()

        on_connection_success_future_dup = Future()

        def on_connection_success_callback_dup(connection, callback_data: OnConnectionSuccessData):
            on_connection_success_future_dup.set_result({})

        # Reuse the same client_id to displace the first client.
        connection_dup = self._create_connection(
            endpoint=test_input_endpoint,
            tls_context=test_tls,
            client_id=connection.client_id,
            on_connection_success_callback=on_connection_success_callback_dup)

        connection_dup.connect().result(TIMEOUT)
        on_connection_success_future_dup.result(TIMEOUT)
        connection_dup.disconnect().result(TIMEOUT)

        # After the second client disconnects, the first one should reconnect,
        # and on_connection_success callback should be fired once again.
        on_connection_resumed_future.result(TIMEOUT)
        success_data = on_connection_success_future.result(TIMEOUT)

        self.assertEqual(success_data['return_code'], ConnectReturnCode.ACCEPTED)
        self.assertEqual(success_data['session_present'], False)

        connection.disconnect().result(TIMEOUT)
        on_connection_closed_future.result(TIMEOUT)

    def test_connect_disconnect_with_callbacks_happy_on_resume(self):
        test_retry_wrapper(self._test_connect_disconnect_with_callbacks_happy_on_resume)

    # ==============================================================
    #             MOSQUITTO CONNECTION TESTS
    # ==============================================================

    def _test_mqtt311_direct_connect_minimum(self):
        input_host_name = _get_env_variable("AWS_TEST_MQTT311_DIRECT_MQTT_HOST")
        input_port = int(_get_env_variable("AWS_TEST_MQTT311_DIRECT_MQTT_PORT"))

        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)
        client = Client(bootstrap, None)
        connection = Connection(
            client=client,
            client_id=create_client_id(),
            host_name=input_host_name,
            port=input_port)
        connection.connect().result(TIMEOUT)
        connection.disconnect().result(TIMEOUT)

    def test_mqtt311_direct_connect_minimum(self):
        test_retry_wrapper(self._test_mqtt311_direct_connect_minimum)

    def _test_mqtt311_direct_connect_basic_auth(self):
        input_host_name = _get_env_variable("AWS_TEST_MQTT311_DIRECT_MQTT_BASIC_AUTH_HOST")
        input_port = int(_get_env_variable("AWS_TEST_MQTT311_DIRECT_MQTT_BASIC_AUTH_PORT"))
        input_username = _get_env_variable("AWS_TEST_MQTT311_BASIC_AUTH_USERNAME")
        input_password = _get_env_variable("AWS_TEST_MQTT311_BASIC_AUTH_PASSWORD")

        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)
        client = Client(bootstrap, None)
        connection = Connection(
            client=client,
            client_id=create_client_id(),
            host_name=input_host_name,
            port=input_port,
            username=input_username,
            password=input_password)
        connection.connect().result(TIMEOUT)
        connection.disconnect().result(TIMEOUT)

    def test_mqtt311_direct_connect_basic_auth(self):
        test_retry_wrapper(self._test_mqtt311_direct_connect_basic_auth)

    def _test_mqtt311_direct_connect_tls(self):
        input_host_name = _get_env_variable("AWS_TEST_MQTT311_DIRECT_MQTT_TLS_HOST")
        input_port = int(_get_env_variable("AWS_TEST_MQTT311_DIRECT_MQTT_TLS_PORT"))

        tls_ctx_options = TlsContextOptions()
        tls_ctx_options.verify_peer = False
        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)
        client = Client(bootstrap, ClientTlsContext(tls_ctx_options))
        connection = Connection(
            client=client,
            client_id=create_client_id(),
            host_name=input_host_name,
            port=input_port)
        connection.connect().result(TIMEOUT)
        connection.disconnect().result(TIMEOUT)

    def test_mqtt311_direct_connect_tls(self):
        test_retry_wrapper(self._test_mqtt311_direct_connect_tls)

    def _test_mqtt311_direct_connect_mutual_tls(self):
        input_cert = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_CERT")
        input_key = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_RSA_KEY")
        input_host = _get_env_variable("AWS_TEST_MQTT311_IOT_CORE_HOST")

        tls_ctx_options = TlsContextOptions.create_client_with_mtls_from_path(
            input_cert,
            input_key
        )
        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)
        client = Client(bootstrap, ClientTlsContext(tls_ctx_options))
        connection = Connection(
            client=client,
            client_id=create_client_id(),
            host_name=input_host,
            port=8883)
        connection.connect().result(TIMEOUT)
        connection.disconnect().result(TIMEOUT)

    def test_mqtt311_direct_connect_mutual_tls(self):
        test_retry_wrapper(self._test_mqtt311_direct_connect_mutual_tls)

    def _test_mqtt311_direct_connect_http_proxy_tls(self):
        input_proxy_host = _get_env_variable("AWS_TEST_MQTT311_PROXY_HOST")
        input_proxy_port = int(_get_env_variable("AWS_TEST_MQTT311_PROXY_PORT"))
        input_host_name = _get_env_variable("AWS_TEST_MQTT311_DIRECT_MQTT_TLS_HOST")
        input_port = int(_get_env_variable("AWS_TEST_MQTT311_DIRECT_MQTT_TLS_PORT"))

        tls_ctx_options = TlsContextOptions()
        tls_ctx_options.verify_peer = False
        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)
        client = Client(bootstrap, ClientTlsContext(tls_ctx_options))

        http_proxy_options = http.HttpProxyOptions(
            host_name=input_proxy_host,
            port=input_proxy_port
        )
        http_proxy_options.connection_type = http.HttpProxyConnectionType.Tunneling
        http_proxy_options.auth_type = http.HttpProxyAuthenticationType.Nothing

        connection = Connection(
            client=client,
            client_id=create_client_id(),
            host_name=input_host_name,
            port=input_port,
            proxy_options=http_proxy_options)
        connection.connect().result(TIMEOUT)
        connection.disconnect().result(TIMEOUT)

    def test_mqtt311_direct_connect_http_proxy_tls(self):
        test_retry_wrapper(self._test_mqtt311_direct_connect_http_proxy_tls)

    def _test_mqtt311_websocket_connect_minimum(self):
        input_host_name = _get_env_variable("AWS_TEST_MQTT311_WS_MQTT_HOST")
        input_port = int(_get_env_variable("AWS_TEST_MQTT311_WS_MQTT_PORT"))

        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)

        def sign_function(transform_args, **kwargs):
            transform_args.set_done()

        client = Client(bootstrap, None)
        connection = Connection(
            client=client,
            client_id=create_client_id(),
            host_name=input_host_name,
            port=input_port,
            use_websockets=True,
            websocket_handshake_transform=sign_function)
        connection.connect().result(TIMEOUT)
        connection.disconnect().result(TIMEOUT)

    def test_mqtt311_websocket_connect_minimum(self):
        test_retry_wrapper(self._test_mqtt311_websocket_connect_minimum)

    def _test_mqtt311_websocket_connect_basic_auth(self):
        input_host_name = _get_env_variable("AWS_TEST_MQTT311_WS_MQTT_BASIC_AUTH_HOST")
        input_port = int(_get_env_variable("AWS_TEST_MQTT311_WS_MQTT_BASIC_AUTH_PORT"))
        input_username = _get_env_variable("AWS_TEST_MQTT311_BASIC_AUTH_USERNAME")
        input_password = _get_env_variable("AWS_TEST_MQTT311_BASIC_AUTH_PASSWORD")

        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)

        def sign_function(transform_args, **kwargs):
            transform_args.set_done()

        client = Client(bootstrap, None)
        connection = Connection(
            client=client,
            client_id=create_client_id(),
            host_name=input_host_name,
            port=input_port,
            username=input_username,
            password=input_password,
            use_websockets=True,
            websocket_handshake_transform=sign_function)
        connection.connect().result(TIMEOUT)
        connection.disconnect().result(TIMEOUT)

    def test_mqtt311_websocket_connect_basic_auth(self):
        test_retry_wrapper(self._test_mqtt311_websocket_connect_basic_auth)

    def _test_mqtt311_websocket_connect_tls(self):
        input_host_name = _get_env_variable("AWS_TEST_MQTT311_WS_MQTT_TLS_HOST")
        input_port = int(_get_env_variable("AWS_TEST_MQTT311_WS_MQTT_TLS_PORT"))

        tls_ctx_options = TlsContextOptions()
        tls_ctx_options.verify_peer = False
        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)

        def sign_function(transform_args, **kwargs):
            transform_args.set_done()

        client = Client(bootstrap, ClientTlsContext(tls_ctx_options))
        connection = Connection(
            client=client,
            client_id=create_client_id(),
            host_name=input_host_name,
            port=input_port,
            use_websockets=True,
            websocket_handshake_transform=sign_function)
        connection.connect().result(TIMEOUT)
        connection.disconnect().result(TIMEOUT)

    def test_mqtt311_websocket_connect_tls(self):
        test_retry_wrapper(self._test_mqtt311_websocket_connect_tls)

    def _test_mqtt311_websocket_connect_http_proxy_tls(self):
        input_proxy_host = _get_env_variable("AWS_TEST_MQTT311_PROXY_HOST")
        input_proxy_port = int(_get_env_variable("AWS_TEST_MQTT311_PROXY_PORT"))
        input_host_name = _get_env_variable("AWS_TEST_MQTT311_WS_MQTT_TLS_HOST")
        input_port = int(_get_env_variable("AWS_TEST_MQTT311_WS_MQTT_TLS_PORT"))

        tls_ctx_options = TlsContextOptions()
        tls_ctx_options.verify_peer = False
        elg = EventLoopGroup()
        resolver = DefaultHostResolver(elg)
        bootstrap = ClientBootstrap(elg, resolver)

        http_proxy_options = http.HttpProxyOptions(
            host_name=input_proxy_host,
            port=input_proxy_port
        )
        http_proxy_options.connection_type = http.HttpProxyConnectionType.Tunneling
        http_proxy_options.auth_type = http.HttpProxyAuthenticationType.Nothing

        def sign_function(transform_args, **kwargs):
            transform_args.set_done()

        client = Client(bootstrap, ClientTlsContext(tls_ctx_options))
        connection = Connection(
            client=client,
            client_id=create_client_id(),
            host_name=input_host_name,
            port=input_port,
            use_websockets=True,
            websocket_handshake_transform=sign_function,
            proxy_options=http_proxy_options)
        connection.connect().result(TIMEOUT)
        connection.disconnect().result(TIMEOUT)

    def test_mqtt311_websocket_connect_http_proxy_tls(self):
        test_retry_wrapper(self._test_mqtt311_websocket_connect_http_proxy_tls)


if __name__ == 'main':
    unittest.main()



"""
HTTP

All network operations in `awscrt.http` are asynchronous.
"""

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0.

import _awscrt
from concurrent.futures import Future
from awscrt import NativeResource
import awscrt.exceptions
from awscrt.io import ClientBootstrap, InputStream, TlsConnectionOptions, SocketOptions
from enum import IntEnum
from typing import List, Tuple, Dict, Optional, Union, Iterator, Callable, Any


class HttpVersion(IntEnum):
    """HTTP protocol version enumeration"""
    Unknown = 0  #: Unknown
    Http1_0 = 1  #: HTTP/1.0
    Http1_1 = 2  #: HTTP/1.1
    Http2 = 3  #: HTTP/2


class Http2SettingID(IntEnum):
    """HTTP/2 Predefined settings(RFC-9113 6.5.2)."""
    HEADER_TABLE_SIZE = 1
    ENABLE_PUSH = 2
    MAX_CONCURRENT_STREAMS = 3
    INITIAL_WINDOW_SIZE = 4
    MAX_FRAME_SIZE = 5
    MAX_HEADER_LIST_SIZE = 6


class Http2Setting:
    """HTTP/2 Setting.
    Settings are very complicated in HTTP/2.
    Each end has its own settings, and the local settings cannot be applied
    until the remote end acknowledges it.
    Each end can change its settings at any time, while the order of the settings
    changed may also result in different behavior.

    Each setting has its boundary and initial values defined in RFC-9113 6.5.2:
    Initial values are listed below, while the range can be found in VALID_RANGES:
    HEADER_TABLE_SIZE: 4096
    ENABLE_PUSH: 1
    MAX_CONCURRENT_STREAMS: 2^32-1
    INITIAL_WINDOW_SIZE: 2^16-1
    MAX_FRAME_SIZE: 2^14
    MAX_HEADER_LIST_SIZE: 2^32-1

    Args:
        id (Http2SettingID): Setting ID.
        value (int): Setting value.
    """
    VALID_RANGES: Dict[Http2SettingID, Tuple[int, int]] = {
        Http2SettingID.HEADER_TABLE_SIZE: (0, 2**32 - 1),
        Http2SettingID.ENABLE_PUSH: (0, 1),
        Http2SettingID.MAX_CONCURRENT_STREAMS: (0, 2**32 - 1),
        Http2SettingID.INITIAL_WINDOW_SIZE: (0, 2**31 - 1),
        Http2SettingID.MAX_FRAME_SIZE: (2**14, 2**24 - 1),
        Http2SettingID.MAX_HEADER_LIST_SIZE: (0, 2**32 - 1),
    }

    def __init__(self, id: Http2SettingID, value: int) -> None:
        assert isinstance(id, Http2SettingID)
        assert isinstance(value, int)
        self.id = id

        # Verify value is within allowed range for the given setting
        self._validate_setting_value(id, value)
        self.value = value

    def _validate_setting_value(self, id: Http2SettingID, value: int) -> None:
        """Validate that setting value is within its allowed range according to RFC-9113 6.5.2."""
        min_value, max_value = self.VALID_RANGES[id]
        if not min_value <= value <= max_value:
            setting_name = id.name
            raise ValueError(f"{setting_name} must be between {min_value} and {max_value}, got {value}")

    def __str__(self) -> str:
        return self.__class__.__name__ + f"({self.id.name}={self.value})"


class HttpConnectionBase(NativeResource):
    """Base for HTTP connection classes."""

    __slots__ = ('_shutdown_future', '_version')

    def __init__(self) -> None:
        super().__init__()
        self._shutdown_future = Future()

    @property
    def shutdown_future(self) -> "concurrent.futures.Future":
        """
        concurrent.futures.Future: Completes when this connection has finished shutting down.
        Future will contain a result of None, or an exception indicating why shutdown occurred.
        Note that the connection may have been garbage-collected before this future completes.
        """
        return self._shutdown_future

    @property
    def version(self) -> HttpVersion:
        """HttpVersion: Protocol used by this connection"""
        return self._version

    def is_open(self) -> bool:
        """
        Returns:
            bool: True if this connection is open and usable, False otherwise.
            Check :attr:`shutdown_future` to know when the connection is completely
            finished shutting down.
        """
        return _awscrt.http_connection_is_open(self._binding)


class HttpClientConnectionBase(HttpConnectionBase):
    __slots__ = ('_host_name', '_port')

    @staticmethod
    def _generic_new(
            host_name: str,
            port: int,
            bootstrap: Optional[ClientBootstrap] = None,
            socket_options: Optional[SocketOptions] = None,
            tls_connection_options: Optional[TlsConnectionOptions] = None,
            proxy_options: Optional['HttpProxyOptions'] = None,
            expected_version: Optional[HttpVersion] = None,
            initial_settings: Optional[List[Http2Setting]] = None,
            on_remote_settings_changed: Optional[Callable[[List[Http2Setting]], None]] = None,
            asyncio_connection=False) -> "concurrent.futures.Future":
        """
        Initialize the generic part of the HttpClientConnection class.
        """
        assert isinstance(bootstrap, ClientBootstrap) or bootstrap is None
        assert isinstance(host_name, str)
        assert isinstance(port, int)
        assert isinstance(tls_connection_options, TlsConnectionOptions) or tls_connection_options is None
        assert isinstance(socket_options, SocketOptions) or socket_options is None
        assert isinstance(proxy_options, HttpProxyOptions) or proxy_options is None

        future = Future()

        try:
            if not socket_options:
                socket_options = SocketOptions()

            if not bootstrap:
                bootstrap = ClientBootstrap.get_or_create_static_default()

            connection_core = _HttpClientConnectionCore(
                host_name,
                port,
                bootstrap=bootstrap,
                tls_connection_options=tls_connection_options,
                connect_future=future,
                expected_version=expected_version,
                on_remote_settings_changed=on_remote_settings_changed,
                asyncio_connection=asyncio_connection)

            _awscrt.http_client_connection_new(
                bootstrap,
                host_name,
                port,
                socket_options,
                tls_connection_options,
                proxy_options,
                initial_settings,
                on_remote_settings_changed,
                connection_core)

        except Exception as e:
            future.set_exception(e)

        return future

    @property
    def host_name(self) -> str:
        """Remote hostname"""
        return self._host_name

    @property
    def port(self) -> int:
        """Remote port"""
        return self._port


class HttpClientConnection(HttpClientConnectionBase):
    """
    An HTTP client connection.

    Use :meth:`HttpClientConnection.new()` to establish a new connection.
    """
    __slots__ = ('_host_name', '_port')

    @classmethod
    def new(cls,
            host_name: str,
            port: int,
            bootstrap: Optional[ClientBootstrap] = None,
            socket_options: Optional[SocketOptions] = None,
            tls_connection_options: Optional[TlsConnectionOptions] = None,
            proxy_options: Optional['HttpProxyOptions'] = None) -> "concurrent.futures.Future":
        """
        Asynchronously establish a new HttpClientConnection.

        Args:
            host_name (str): Connect to host.

            port (int): Connect to port.

            bootstrap (Optional [ClientBootstrap]): Client bootstrap to use when initiating socket connection.
                If None is provided, the default singleton is used.

            socket_options (Optional[SocketOptions]): Optional socket options.
                If None is provided, then default options are used.

            tls_connection_options (Optional[TlsConnectionOptions]): Optional TLS
                connection options. If None is provided, then the connection will
                be attempted over plain-text.

            proxy_options (Optional[HttpProxyOptions]): Optional proxy options.
                If None is provided then a proxy is not used.

        Returns:
            concurrent.futures.Future: A Future which completes when connection succeeds or fails.
            If successful, the Future will contain a new :class:`HttpClientConnection`.
            Otherwise, it will contain an exception.
        """
        return cls._generic_new(
            host_name,
            port,
            bootstrap,
            socket_options,
            tls_connection_options,
            proxy_options)

    def request(self,
                request: 'HttpRequest',
                on_response: Optional[Callable[..., None]] = None,
                on_body: Optional[Callable[..., None]] = None) -> 'HttpClientStream':
        """Create :class:`HttpClientStream` to carry out the request/response exchange.

        NOTE: The HTTP stream sends no data until :meth:`HttpClientStream.activate()`
        is called. Call activate() when you're ready for callbacks and events to fire.

        Args:
            request (HttpRequest): Definition for outgoing request.

            on_response: Optional callback invoked once main response headers are received.
                The function should take the following arguments and return nothing:

                    *   `http_stream` (:class:`HttpClientStream`): HTTP stream carrying
                        out this request/response exchange.

                    *   `status_code` (int): Response status code.

                    *   `headers` (List[Tuple[str, str]]): Response headers as a
                        list of (name,value) pairs.

                    *   `**kwargs` (dict): Forward compatibility kwargs.

                An exception raise by this function will cause the HTTP stream to end in error.
                This callback is always invoked on the connection's event-loop thread.

            on_body: Optional callback invoked 0+ times as response body data is received.
                The function should take the following arguments and return nothing:

                    *   `http_stream` (:class:`HttpClientStream`): HTTP stream carrying
                        out this request/response exchange.

                    *   `chunk` (buffer): Response body data (not necessarily
                        a whole "chunk" of chunked encoding).

                    *   `**kwargs` (dict): Forward-compatibility kwargs.

                An exception raise by this function will cause the HTTP stream to end in error.
                This callback is always invoked on the connection's event-loop thread.

        Returns:
            HttpClientStream:
        """
        return HttpClientStream(self, request, on_response, on_body)

    def close(self) -> "concurrent.futures.Future":
        """Close the connection.

        Shutdown is asynchronous. This call has no effect if the connection is already
        closing.

        Returns:
            concurrent.futures.Future: This connection's :attr:`shutdown_future`,
            which completes when shutdown has finished.
        """
        _awscrt.http_connection_close(self._binding)
        return self.shutdown_future


class Http2ClientConnection(HttpClientConnectionBase):

    @classmethod
    def new(cls,
            host_name: str,
            port: int,
            bootstrap: Optional[ClientBootstrap] = None,
            socket_options: Optional[SocketOptions] = None,
            tls_connection_options: Optional[TlsConnectionOptions] = None,
            proxy_options: Optional['HttpProxyOptions'] = None,
            initial_settings: Optional[List[Http2Setting]] = None,
            on_remote_settings_changed: Optional[Callable[[List[Http2Setting]],
                                                          None]] = None) -> "concurrent.futures.Future":
        """
        Asynchronously establish an HTTP/2 client connection.
        Notes: to set up the connection, the server must support HTTP/2 and TlsConnectionOptions

        This class extends HttpClientConnection with HTTP/2 specific functionality.

        HTTP/2 specific args:
            initial_settings (List[Http2Setting]): The initial settings to change for the connection.

            on_remote_settings_changed: Optional callback invoked once the remote peer changes its settings.
                And the settings are acknowledged by the local connection.
                The function should take the following arguments and return nothing:

                    *   `settings` (List[Http2Setting]): List of settings that were changed.
        """
        return cls._generic_new(
            host_name,
            port,
            bootstrap,
            socket_options,
            tls_connection_options,
            proxy_options,
            HttpVersion.Http2,
            initial_settings,
            on_remote_settings_changed)

    def request(self,
                request: 'HttpRequest',
                on_response: Optional[Callable[..., None]] = None,
                on_body: Optional[Callable[..., None]] = None,
                manual_write: bool = False) -> 'Http2ClientStream':
        """Create `Http2ClientStream` to carry out the request/response exchange.

        NOTE: The HTTP stream sends no data until `Http2ClientStream.activate()`
        is called. Call activate() when you're ready for callbacks and events to fire.

        Args:
            request (HttpRequest): Definition for outgoing request.

            on_response: Optional callback invoked once main response headers are received.
                The function should take the following arguments and return nothing:

                    *   `http_stream` (`Http2ClientStream`): HTTP/2 stream carrying
                        out this request/response exchange.

                    *   `status_code` (int): Response status code.

                    *   `headers` (List[Tuple[str, str]]): Response headers as a
                        list of (name,value) pairs.

                    *   `**kwargs` (dict): Forward compatibility kwargs.

            on_body: Optional callback invoked 0+ times as response body data is received.
                The function should take the following arguments and return nothing:

                    *   `http_stream` (`Http2ClientStream`): HTTP/2 stream carrying
                        out this request/response exchange.

                    *   `chunk` (buffer): Response body data (not necessarily
                        a whole "chunk" of chunked encoding).

                    *   `**kwargs` (dict): Forward-compatibility kwargs.

            manual_write (bool): If True, enables manual data writing on the stream.
                This allows calling `write_data()` to stream the request body in chunks.
                Note: In the asyncio version, this is replaced by the async_body parameter.

        Returns:
            Http2ClientStream: Stream for the HTTP/2 request/response exchange.
        """
        return Http2ClientStream(self, request, on_response, on_body, manual_write)

    def close(self) -> "concurrent.futures.Future":
        """Close the connection.

        Shutdown is asynchronous. This call has no effect if the connection is already
        closing.

        Returns:
            concurrent.futures.Future: This connection's :attr:`shutdown_future`,
            which completes when shutdown has finished.
        """
        _awscrt.http_connection_close(self._binding)
        return self.shutdown_future


class HttpStreamBase(NativeResource):
    """Base for HTTP stream classes.

    Attributes:
        connection: The HTTP connection this stream belongs to.
        completion_future: Future that completes when the operation finishes.
    """
    __slots__ = ('_connection', '_completion_future', '_on_body_cb')

    def __init__(self, connection, on_body: Optional[Callable[..., None]] = None) -> None:
        super().__init__()
        self._connection = connection
        self._completion_future = Future()
        self._on_body_cb: Optional[Callable[..., None]] = on_body

    @property
    def connection(self) -> HttpConnectionBase:
        return self._connection

    @property
    def completion_future(self) -> "concurrent.futures.Future":
        return self._completion_future

    def _on_body(self, chunk: bytes) -> None:
        if self._on_body_cb:
            self._on_body_cb(http_stream=self, chunk=chunk)


class HttpClientStreamBase(HttpStreamBase):
    """Base for HTTP client stream classes.

    Attributes:
        connection: This stream's connection.

        completion_future: Future that completes when
            the request/response exchange is finished.
    """

    __slots__ = ('_response_status_code', '_on_response_cb', '_on_body_cb', '_request', '_version')

    def _init_common(self,
                     connection: HttpClientConnectionBase,
                     request: 'HttpRequest',
                     on_response: Optional[Callable[..., None]] = None,
                     on_body: Optional[Callable[..., None]] = None,
                     http2_manual_write: bool = False) -> None:
        assert isinstance(connection, HttpClientConnectionBase)
        assert isinstance(request, HttpRequest)
        assert callable(on_response) or on_response is None
        assert callable(on_body) or on_body is None

        super().__init__(connection, on_body)

        self._on_response_cb: Optional[Callable[..., None]] = on_response
        self._response_status_code: Optional[int] = None

        # keep HttpRequest alive until stream completes
        self._request = request
        self._version = connection.version
        self._binding = _awscrt.http_client_stream_new(self, connection, request, http2_manual_write)

    @property
    def version(self) -> HttpVersion:
        """HttpVersion: Protocol used by this stream"""
        return self._version

    @property
    def response_status_code(self) -> Optional[int]:
        """int: The response status code.

        This is None until a response arrives."""
        return self._response_status_code

    def _on_response(self, status_code: int, name_value_pairs: List[Tuple[str, str]]) -> None:
        self._response_status_code = status_code

        if self._on_response_cb:
            self._on_response_cb(http_stream=self, status_code=status_code, headers=name_value_pairs)

    def _on_complete(self, error_code: int) -> None:
        # done with HttpRequest, drop reference
        self._request = None  # type: ignore

        if error_code == 0:
            self._completion_future.set_result(self._response_status_code)
        else:
            self._completion_future.set_exception(awscrt.exceptions.from_code(error_code))


class HttpClientStream(HttpClientStreamBase):
    """HTTP stream that sends a request and receives a response.

    Create an HttpClientStream with :meth:`HttpClientConnection.request()`.

    NOTE: The HTTP stream sends no data until :meth:`HttpClientStream.activate()`
    is called. Call activate() when you're ready for callbacks and events to fire.

    Attributes:
        connection (HttpClientConnection): This stream's connection.

        completion_future (concurrent.futures.Future): Future that will contain
            the response status code (int) when the request/response exchange
            completes. If the exchange fails to complete, the Future will
            contain an exception indicating why it failed.
    """

    def __init__(self,
                 connection: HttpClientConnection,
                 request: 'HttpRequest',
                 on_response: Optional[Callable[..., None]] = None,
                 on_body: Optional[Callable[..., None]] = None) -> None:
        self._init_common(connection, request, on_response, on_body)

    def activate(self) -> None:
        """Begin sending the request.

        The HTTP stream does nothing until this is called. Call activate() when you
        are ready for its callbacks and events to fire.
        """
        _awscrt.http_client_stream_activate(self)


class Http2ClientStream(HttpClientStreamBase):
    def __init__(self,
                 connection: HttpClientConnection,
                 request: 'HttpRequest',
                 on_response: Optional[Callable[..., None]] = None,
                 on_body: Optional[Callable[..., None]] = None,
                 manual_write: bool = False) -> None:
        self._init_common(connection, request, on_response, on_body, manual_write)

    def activate(self) -> None:
        """Begin sending the request.

        The HTTP stream does nothing until this is called. Call activate() when you
        are ready for its callbacks and events to fire.
        """
        _awscrt.http_client_stream_activate(self)

    def write_data(self,
                   data_stream: Union[InputStream, Any],
                   end_stream: bool = False) -> "concurrent.futures.Future":
        """Write a chunk of data to the request body stream.

        This method is only available when the stream was created with
        manual_write=True. This allows incremental writing of request data.

        Note: In the asyncio version, this is replaced by the request_body_generator parameter
        which accepts an async generator.

        Args:
            data_stream (Union[InputStream, Any]): Data to write. If not an InputStream,
                it will be wrapped in one. Can be None to send an empty chunk.

            end_stream (bool): True to indicate this is the last chunk and no more data
                will be sent. False if more chunks will follow.

        Returns:
            concurrent.futures.Future: Future that completes when the write operation
                is done. The future will contain None on success, or an exception on failure.
        """
        future = Future()
        body_stream = InputStream.wrap(data_stream, allow_none=True)

        def on_write_complete(error_code: int) -> None:
            if future.cancelled():
                # the future was cancelled, so we don't need to set the result or exception
                return
            if error_code:
                future.set_exception(awscrt.exceptions.from_code(error_code))
            else:
                future.set_result(None)

        _awscrt.http2_client_stream_write_data(self, body_stream, end_stream, on_write_complete)
        return future


class HttpMessageBase(NativeResource):
    """
    Base for HttpRequest and HttpResponse classes.
    """
    __slots__ = ('_headers', '_body_stream')

    def __init__(self, binding: Any, headers: 'HttpHeaders',
                 body_stream: Optional[Union[InputStream, Any]] = None) -> None:
        assert isinstance(headers, HttpHeaders)

        super().__init__()
        self._binding = binding
        self._headers = headers
        self._body_stream: Optional[InputStream] = None

        if body_stream:
            self.body_stream = body_stream

    @property
    def headers(self) -> 'HttpHeaders':
        """HttpHeaders: Headers to send."""
        return self._headers

    @property
    def body_stream(self) -> Optional[InputStream]:
        return self._body_stream

    @body_stream.setter
    def body_stream(self, stream: Union[InputStream, Any]) -> None:
        self._body_stream = InputStream.wrap(stream)
        _awscrt.http_message_set_body_stream(self._binding, self._body_stream)


class HttpRequest(HttpMessageBase):
    """
    Definition for an outgoing HTTP request.

    The request may be transformed (ex: signing the request) before its data is eventually sent.

    Args:
        method (str): HTTP request method (verb). Default value is "GET".
        path (str): HTTP path-and-query value. Default value is "/".
        headers (Optional[HttpHeaders]): Optional headers. If None specified,
            an empty :class:`HttpHeaders` is created.
        body_stream(Optional[Union[InputStream, io.IOBase]]): Optional body as binary stream.
    """

    __slots__ = ()

    def __init__(self,
                 method: str = 'GET',
                 path: str = '/',
                 headers: Optional['HttpHeaders'] = None,
                 body_stream: Optional[Union[InputStream, Any]] = None) -> None:
        assert isinstance(headers, HttpHeaders) or headers is None

        if headers is None:
            headers = HttpHeaders()

        binding = _awscrt.http_message_new_request(headers)
        super().__init__(binding, headers, body_stream)
        self.method = method
        self.path = path

    @classmethod
    def _from_bindings(cls, request_binding: Any, headers_binding: Any) -> 'HttpRequest':
        """Construct HttpRequest and its HttpHeaders from pre-existing native objects"""

        # avoid class's default constructor
        # just invoke parent class's __init__()
        request = cls.__new__(cls)
        headers = HttpHeaders._from_binding(headers_binding)
        super(cls, request).__init__(request_binding, headers)
        return request

    @property
    def method(self) -> str:
        """str: HTTP request method (verb)."""
        return _awscrt.http_message_get_request_method(self._binding)

    @method.setter
    def method(self, method: str) -> None:
        _awscrt.http_message_set_request_method(self._binding, method)

    @property
    def path(self) -> str:
        """str: HTTP path-and-query value."""
        return _awscrt.http_message_get_request_path(self._binding)

    @path.setter
    def path(self, path: str) -> None:
        return _awscrt.http_message_set_request_path(self._binding, path)


class HttpHeaders(NativeResource):
    """
    Collection of HTTP headers.

    A given header name may have multiple values.
    Header names are always treated in a case-insensitive manner.
    HttpHeaders can be iterated over as (name,value) pairs.

    Args:
        name_value_pairs (Optional[List[Tuple[str, str]]]): Construct from a
            collection of (name,value) pairs.
    """

    __slots__ = ()

    def __init__(self, name_value_pairs: Optional[List[Tuple[str, str]]] = None) -> None:
        super().__init__()
        self._binding = _awscrt.http_headers_new()
        if name_value_pairs:
            self.add_pairs(name_value_pairs)

    @classmethod
    def _from_binding(cls, binding: Any) -> 'HttpHeaders':
        """Construct from a pre-existing native object"""
        headers = cls.__new__(cls)  # avoid class's default constructor
        super(cls, headers).__init__()  # just invoke parent class's __init__()
        headers._binding = binding
        return headers

    def add(self, name: str, value: str) -> None:
        """
        Add a name-value pair.

        Args:
            name (str): Name.
            value (str): Value.
        """
        assert isinstance(name, str)
        assert isinstance(value, str)
        _awscrt.http_headers_add(self._binding, name, value)

    def add_pairs(self, name_value_pairs: List[Tuple[str, str]]) -> None:
        """
        Add list of (name,value) pairs.

        Args:
            name_value_pairs (List[Tuple[str, str]]): List of (name,value) pairs.
        """
        _awscrt.http_headers_add_pairs(self._binding, name_value_pairs)

    def set(self, name: str, value: str) -> None:
        """
        Set a name-value pair, any existing values for the name are removed.

        Args:
            name (str): Name.
            value (str): Value.
        """
        assert isinstance(name, str)
        assert isinstance(value, str)
        _awscrt.http_headers_set(self._binding, name, value)

    def get_values(self, name: str) -> Iterator[str]:
        """
        Return an iterator over the values for this name.

        Args:
            name (str): Name.

        Returns:
            Iterator[str]: Iterator over values for this header name
        """
        assert isinstance(name, str)
        name = name.lower()
        for i in range(_awscrt.http_headers_count(self._binding)):
            name_i, value_i = _awscrt.http_headers_get_index(self._binding, i)
            if name_i.lower() == name:
                yield value_i

    def get(self, name: str, default: Optional[str] = None) -> Optional[str]:
        """
        Get the first value for this name, ignoring any additional values.
        Returns `default` if no values exist.

        Args:
            name (str): Name.
            default (Optional[str]): If `name` not found, this value is returned.
                Defaults to None.
        Returns:
            Optional[str]: Header value or default
        """
        assert isinstance(name, str)
        return _awscrt.http_headers_get(self._binding, name, default)

    def remove(self, name: str) -> None:
        """
        Remove all values for this name.
        Raises a KeyError if name not found.

        Args:
            name (str): Header name.
        """
        assert isinstance(name, str)
        _awscrt.http_headers_remove(self._binding, name)

    def remove_value(self, name: str, value: str) -> None:
        """
        Remove a specific value for this name.
        Raises a ValueError if value not found.

        Args:
            name (str): Name.
            value (str): Value.
        """
        assert isinstance(name, str)
        assert isinstance(value, str)
        _awscrt.http_headers_remove_value(self._binding, name, value)

    def clear(self) -> None:
        """
        Clear all headers.
        """
        _awscrt.http_headers_clear(self._binding)

    def __iter__(self) -> Iterator[Tuple[str, str]]:
        """
        Iterate over all (name,value) pairs.
        """
        for i in range(_awscrt.http_headers_count(self._binding)):
            yield _awscrt.http_headers_get_index(self._binding, i)

    def __str__(self) -> str:
        return self.__class__.__name__ + "(" + str([pair for pair in self]) + ")"


class HttpProxyConnectionType(IntEnum):
    """Proxy connection type enumeration"""
    Legacy = 0
    """
    Use the old connection establishment logic that would use:

         1. Forwarding if not using TLS
         2. Tunneling if using TLS
    """

    Forwarding = 1
    """
    Establish a request forwarding connection to the proxy.

    In this case, TLS is not a valid option.
    """

    Tunneling = 2
    """Establish a tunneling connection through the proxy to the ultimate endpoint."""


class HttpProxyAuthenticationType(IntEnum):
    """Proxy authentication type enumeration."""
    Nothing = 0
    """No authentication"""

    Basic = 1
    """Username and password"""


class HttpProxyOptions:
    """
    Proxy options for HTTP clients.

    Args:
        host_name (str): Name of the proxy server to connect through.

        port (int): Port number of the proxy server to connect through.

        tls_connection_options (Optional[TlsConnectionOptions]): Optional
            `TlsConnectionOptions` for the Local to Proxy connection.
            Must be distinct from the `TlsConnectionOptions`
            provided to the HTTP connection.

        auth_type (HttpProxyAuthenticationType): Type of proxy authentication to use.
            Default is :const:`HttpProxyAuthenticationType.Nothing`.

        auth_username (Optional[str]): Username to use when `auth_type` is
            :const:`HttpProxyAuthenticationType.Basic`.

        auth_password (Optional[str]): Username to use when `auth_type` is
            :const:`HttpProxyAuthenticationType.Basic`.

        connection_type (Optional[HttpProxyConnectionType): Type of proxy connection to make.
            Default is :const:`HttpProxyConnectionType.Legacy`.


    Attributes:
        host_name (str): Name of the proxy server to connect through.

        port (int): Port number of the proxy server to connect through.

        tls_connection_options (Optional[TlsConnectionOptions]): Optional
            `TlsConnectionOptions` for the Local to Proxy connection.
            Must be distinct from the `TlsConnectionOptions`
            provided to the HTTP connection.

        auth_type (HttpProxyAuthenticationType): Type of proxy authentication to use.

        auth_username (Optional[str]): Username to use when `auth_type` is
            :const:`HttpProxyAuthenticationType.Basic`.

        auth_password (Optional[str]): Username to use when `auth_type` is
            :const:`HttpProxyAuthenticationType.Basic`.

        connection_type (HttpProxyConnectionType): Type of proxy connection to make.

    """

    def __init__(self,
                 host_name: str,
                 port: int,
                 tls_connection_options: Optional[TlsConnectionOptions] = None,
                 auth_type: HttpProxyAuthenticationType = HttpProxyAuthenticationType.Nothing,
                 auth_username: Optional[str] = None,
                 auth_password: Optional[str] = None,
                 connection_type: HttpProxyConnectionType = HttpProxyConnectionType.Legacy) -> None:
        self.host_name = host_name
        self.port = port
        self.tls_connection_options = tls_connection_options
        self.auth_type = auth_type
        self.auth_username = auth_username
        self.auth_password = auth_password
        self.connection_type = connection_type


class _HttpClientConnectionCore:
    '''
    Private class to keep all the related Python object alive until C land clean up for HttpClientConnection
    '''

    def __init__(
            self,
            host_name: str,
            port: int,
            bootstrap: Optional[ClientBootstrap] = None,
            tls_connection_options: Optional[TlsConnectionOptions] = None,
            connect_future: Optional[Future] = None,
            expected_version: Optional[HttpVersion] = None,
            on_remote_settings_changed: Optional[Callable[[List[Http2Setting]], None]] = None,
            asyncio_connection=False) -> None:
        self._shutdown_future = None
        self._host_name = host_name
        self._port = port
        self._bootstrap = bootstrap
        self._tls_connection_options = tls_connection_options
        self._connect_future = connect_future
        self._expected_version = expected_version
        self._on_remote_settings_changed_from_user = on_remote_settings_changed
        self._asyncio_connection = asyncio_connection

    def _on_connection_setup(self, binding: Any, error_code: int, http_version: HttpVersion) -> None:
        if self._connect_future is None:
            return
        if error_code != 0:
            self._connect_future.set_exception(awscrt.exceptions.from_code(error_code))
            return
        if self._expected_version and self._expected_version != http_version:
            # unexpected protocol version
            # AWS_ERROR_HTTP_UNSUPPORTED_PROTOCOL
            self._connect_future.set_exception(awscrt.exceptions.from_code(2060))
            return
        if self._asyncio_connection:
            # Import is done here to avoid circular import issues
            from awscrt.aio.http import AIOHttpClientConnection, AIOHttp2ClientConnection
            if http_version == HttpVersion.Http2:
                connection = AIOHttp2ClientConnection()
            else:
                connection = AIOHttpClientConnection()
        else:
            if http_version == HttpVersion.Http2:
                connection = Http2ClientConnection()
            else:
                connection = HttpClientConnection()

        connection._host_name = self._host_name
        connection._port = self._port

        connection._binding = binding
        connection._version = HttpVersion(http_version)
        self._shutdown_future = connection.shutdown_future
        self._connect_future.set_result(connection)
        # release reference to the future, as it points to connection which creates a cycle reference.
        self._connect_future = None

    def _on_shutdown(self, error_code: int) -> None:
        if self._shutdown_future is None:
            # connection failed, ignore shutdown
            return
        if error_code:
            self._shutdown_future.set_exception(awscrt.exceptions.from_code(error_code))
        else:
            self._shutdown_future.set_result(None)

    def _on_remote_settings_changed(self, native_settings: List[Tuple[int, int]]) -> None:
        if self._on_remote_settings_changed_from_user:
            # convert the list of tuple to list of Http2Setting
            settings = [Http2Setting(Http2SettingID(id), value) for id, value in native_settings]
            self._on_remote_settings_changed_from_user(settings)



"""
Cross-platform library for `awscrt`.
"""
import _awscrt


def get_cpu_group_count() -> int:
    """
    Returns number of processor groups on the system.

    Useful for working with non-uniform memory access (NUMA) nodes.
    """
    return _awscrt.get_cpu_group_count()


def get_cpu_count_for_group(group_idx: int) -> int:
    """
    Returns number of processors in a given group.
    """
    return _awscrt.get_cpu_count_for_group(group_idx)


def join_all_native_threads(*, timeout_sec: float = -1.0) -> bool:
    """
    Waits for all native threads to complete their join call.

    This can only be safely called from the main thread.
    This call may be required for native memory usage to reach zero.

    Args:
        timeout_sec (float): Number of seconds to wait before a timeout exception is raised.
            By default the wait is unbounded.

    Returns:
        bool: Returns whether threads could be joined before the timeout.
    """
    return _awscrt.thread_join_all_managed(timeout_sec)



#!/usr/bin/env python3
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: MIT-0


import logging
import pathlib
import shutil
import subprocess


_TOOLS = [
    "cadical",
    "cbmc",
    "cbmc-viewer",
    "cbmc-starter-kit-update",
    "kissat",
    "litani",
]


def _format_versions(table):
    lines = [
        "<table>",
        '<tr><td colspan="2" style="font-weight: bold">Tool Versions</td></tr>',
    ]
    for tool, version in table.items():
        if version:
            v_str = f'<code><pre style="margin: 0">{version}</pre></code>'
        else:
            v_str = '<em>not found</em>'
        lines.append(
            f'<tr><td style="font-weight: bold; padding-right: 1em; '
            f'text-align: right;">{tool}:</td>'
            f'<td>{v_str}</td></tr>')
    lines.append("</table>")
    return "\n".join(lines)


def _get_tool_versions():
    ret = {}
    for tool in _TOOLS:
        err = f"Could not determine version of {tool}: "
        ret[tool] = None
        if not shutil.which(tool):
            logging.error("%s'%s' not found on $PATH", err, tool)
            continue
        cmd = [tool, "--version"]
        proc = subprocess.Popen(cmd, text=True, stdout=subprocess.PIPE)
        try:
            out, _ = proc.communicate(timeout=10)
        except subprocess.TimeoutExpired:
            logging.error("%s'%s --version' timed out", err, tool)
            continue
        if proc.returncode:
            logging.error(
                "%s'%s --version' returned %s", err, tool, str(proc.returncode))
            continue
        ret[tool] = out.strip()
    return ret


def main():
    exe_name = pathlib.Path(__file__).name
    logging.basicConfig(format=f"{exe_name}: %(message)s")

    table = _get_tool_versions()
    out = _format_versions(table)
    print(out)


if __name__ == "__main__":
    main()



#! /usr/bin/python

import sys
import re

args = sys.argv[1:]
assert len(args) == 2, "usage is <bpl-file-name> <trace-to-convert>"
print args


def make_bpl_c_mapping(bpl_filename):
    linenum = 0
    bpl_c_mapping = {}
    current_bb_line = -1
    with open(bpl_filename) as bpl_file:
        for line in bpl_file:
            linenum += 1
            if re.match(r'\$bb\d+', line):
                current_bb_line = linenum
                bpl_c_mapping[current_bb_line] = []

            matchObj = re.match(r'  assume {:sourceloc "(.+)", (\d+), (\d+)} true;', line)
            if matchObj:
                matched_filename = matchObj.group(1)
                matched_fileline = matchObj.group(2)
                bpl_c_mapping[current_bb_line].append((matched_filename, matched_fileline))

    return bpl_c_mapping


def convert_trace_to_c(trace_filename, bpl_c_mapping):
    with open(trace_filename) as tracefile:
        for line in tracefile:
            matchObj = re.match(r'.*\.bpl\((\d+),\d+\).*', line)
            if matchObj:
                bpl_line = int(matchObj.group(1))
                if bpl_line in bpl_c_mapping:
                    mapping = bpl_c_mapping[bpl_line]
                    if len(mapping) > 0:
                        print mapping[0]
                    else:
                        print "No Source Loc"
                else:
                    print "unknown basic block"

                print "\t\t", matchObj.group(0), matchObj.group(1)


bpl_c_mapping = make_bpl_c_mapping(args[0])
convert_trace_to_c(args[1], bpl_c_mapping)



import Builder
import argparse
from pathlib import Path
import sys


class AWSCrtPython(Builder.Action):

    def run(self, env):
        # allow custom python to be used
        parser = argparse.ArgumentParser()
        parser.add_argument('--python')
        args = parser.parse_known_args(env.args.args)[0]
        if args.python:
            self.python = args.python
        else:
            # Fall back on using the "{python}" builder variable
            self.python = env.config['variables']['python']

        # Create a virtual environment and use that.
        # Otherwise, in places like ubuntu 24.04, PEP 668 stops
        # you from globally installing/upgrading packages
        venv_dirpath = Path.cwd() / '.venv-builder'
        env.shell.exec(self.python, '-m', 'venv', str(venv_dirpath), check=True)
        if sys.platform == 'win32':
            self.python = str(venv_dirpath / 'Scripts/python')
        else:
            self.python = str(venv_dirpath / 'bin/python')

        # Enable S3 tests
        env.shell.setenv('AWS_TEST_S3', '1')
        env.shell.setenv('AWS_TEST_LOCALHOST', '1')

        actions = [
            [self.python, '--version'],
            [self.python, '-m', 'pip', 'install', '--upgrade', 'pip'],
            Builder.SetupCrossCICrtEnvironment(),
            [self.python, '-m', 'pip', 'install', '--verbose', '.[dev]'],
            # "--failfast" because, given how our leak-detection in tests currently works,
            # once one test fails all the rest usually fail too.
            [self.python, '-m', 'unittest', 'discover', '--verbose', '--failfast'],
            # postman-echo.com in now requires TLS1.3,
            # but our Mac implementation doesn't support TLS1.3 yet.
            # The work has been planned to Dec. 2025 to support TLS1.3,
            # so disable the test for now. And reenable it afterward
            # # http_client_test.py launches external processes using the extra args
            # [self.python, 'crt/aws-c-http/integration-testing/http_client_test.py',
            #     self.python, 'elasticurl.py'],
        ]

        return Builder.Script(actions, name='aws-crt-python')



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0.

from awscrt.io import *
from test import NativeResourceTest, TIMEOUT
import io
import os
import sys
import unittest


class EventLoopGroupTest(NativeResourceTest):
    def test_init_defaults(self):
        event_loop_group = EventLoopGroup()

    def test_1_thread(self):
        event_loop_group = EventLoopGroup(1)

    def test_cpu_group(self):
        event_loop_group = EventLoopGroup(cpu_group=0)

    def test_shutdown_complete(self):
        event_loop_group = EventLoopGroup()
        shutdown_event = event_loop_group.shutdown_event
        del event_loop_group
        self.assertTrue(shutdown_event.wait(TIMEOUT))

    def test_init_defaults_singleton(self):
        event_loop_group = EventLoopGroup.get_or_create_static_default()

    def test_init_defaults_singleton_is_singleton(self):
        event_loop_group_one = EventLoopGroup.get_or_create_static_default()
        event_loop_group_two = EventLoopGroup.get_or_create_static_default()
        self.assertTrue(event_loop_group_one == event_loop_group_two)

    def test_shutdown_complete_singleton(self):
        event_loop_group = EventLoopGroup.get_or_create_static_default()
        shutdown_event = event_loop_group.shutdown_event
        del event_loop_group
        EventLoopGroup.release_static_default()
        self.assertTrue(shutdown_event.wait(TIMEOUT))


class DefaultHostResolverTest(NativeResourceTest):
    def test_init(self):
        event_loop_group = EventLoopGroup()
        host_resolver = DefaultHostResolver(event_loop_group)

    def test_init_singleton(self):
        host_resolver = DefaultHostResolver.get_or_create_static_default()

    def test_init_singleton_is_singleton(self):
        host_resolver_one = DefaultHostResolver.get_or_create_static_default()
        host_resolver_two = DefaultHostResolver.get_or_create_static_default()
        self.assertTrue(host_resolver_one == host_resolver_two)


class ClientBootstrapTest(NativeResourceTest):
    def test_create_destroy(self):
        event_loop_group = EventLoopGroup()
        host_resolver = DefaultHostResolver(event_loop_group)
        bootstrap = ClientBootstrap(event_loop_group, host_resolver)

        # ensure shutdown_event fires
        bootstrap_shutdown_event = bootstrap.shutdown_event
        del bootstrap
        self.assertTrue(bootstrap_shutdown_event.wait(TIMEOUT))

    def test_create_destroy_singleton(self):
        bootstrap = ClientBootstrap.get_or_create_static_default()

        # ensure shutdown_event fires
        bootstrap_shutdown_event = bootstrap.shutdown_event
        del bootstrap
        ClientBootstrap.release_static_default()
        self.assertTrue(bootstrap_shutdown_event.wait(TIMEOUT))

    def test_init_singleton_is_singleton(self):
        client_one = ClientBootstrap.get_or_create_static_default()
        client_two = ClientBootstrap.get_or_create_static_default()
        self.assertTrue(client_one == client_two)


class ClientTlsContextTest(NativeResourceTest):
    def test_init_defaults(self):
        opt = TlsContextOptions()
        ctx = ClientTlsContext(opt)

    def test_with_mtls_from_path(self):
        opt = TlsContextOptions.create_client_with_mtls_from_path(
            'test/resources/unittest.crt', 'test/resources/unittest.key')
        ctx = ClientTlsContext(opt)

    def test_with_mtls_pkcs12(self):
        try:
            opt = TlsContextOptions.create_client_with_mtls_pkcs12(
                'test/resources/unittest.p12', '1234')
            ctx = ClientTlsContext(opt)
        except Exception as e:
            if 'PLATFORM_NOT_SUPPORTED' in str(e):
                raise unittest.SkipTest(f'PKCS#12 not supported on this platform ({sys.platform})')
            else:
                # well then this is a real error
                raise e

    def test_override_default_trust_store_dir(self):
        opt = TlsContextOptions()
        opt.override_default_trust_store_from_path('test/resources', None)
        ctx = ClientTlsContext(opt)

    def test_override_default_trust_store_file(self):
        opt = TlsContextOptions()
        opt.override_default_trust_store_from_path(None, 'test/resources/ca.crt')
        ctx = ClientTlsContext(opt)

    def test_set_cipher_preference_tlsv1_2_2025(self):
        opt = TlsContextOptions()
        opt.cipher_pref = TlsCipherPref.TLSv1_2_2025_07

        try:
            ctx = ClientTlsContext(opt)
        except Exception as e:
            if sys.platform.startswith("linux"):
                # On Linux, this should not fail
                self.fail(f"Unexpected error on Linux: {e}")
            else:
                # On non-Linux platforms, verify we get the expected error and skip
                self.assertIn('AWS_IO_TLS_CIPHER_PREF_UNSUPPORTED', str(e))
                self.skipTest(f"TLSv1_2_2025_07 not supported on {sys.platform}")


class TlsConnectionOptionsTest(NativeResourceTest):
    def test_init(self):
        opt = TlsContextOptions()
        ctx = ClientTlsContext(opt)
        conn_opt = TlsConnectionOptions(ctx)

    def test_alpn_list(self):
        opt = TlsContextOptions()
        ctx = ClientTlsContext(opt)
        conn_opt = TlsConnectionOptions(ctx)
        conn_opt.set_alpn_list(['h2', 'http/1.1'])

    def test_server_name(self):
        opt = TlsContextOptions()
        ctx = ClientTlsContext(opt)
        conn_opt = TlsConnectionOptions(ctx)
        conn_opt.set_server_name('localhost')


class MockPythonStream:
    """For testing duck-typed stream classes.
    Doesn't inherit from io.IOBase. Doesn't implement readinto()"""

    def __init__(self, src_data):
        self.data = bytes(src_data)
        self.len = len(src_data)
        self.pos = 0

    def seek(self, where):
        self.pos = where

    def tell(self):
        return self.pos

    def read(self, amount=None):
        if amount is None:
            amount = self.len - self.pos
        else:
            amount = min(amount, self.len - self.pos)
        prev_pos = self.pos
        self.pos += amount
        return self.data[prev_pos: self.pos]


class InputStreamTest(NativeResourceTest):
    def _test(self, python_stream, expected):
        input_stream = InputStream(python_stream)
        result = bytearray()
        fixed_mv_len = 4
        fixed_mv = memoryview(bytearray(fixed_mv_len))
        while True:
            read_len = input_stream._read_into_memoryview(fixed_mv)
            if read_len is None:
                continue
            if read_len == 0:
                break
            if read_len > 0:
                self.assertLessEqual(read_len, fixed_mv_len)
                result += fixed_mv[0:read_len]

        self.assertEqual(expected, result)

    def test_read_official_io(self):
        # Read from a class defined in the io module
        src_data = b'a long string here'
        python_stream = io.BytesIO(src_data)
        self._test(python_stream, src_data)

    def test_read_duck_typed_io(self):
        # Read from a class defined in the io module
        src_data = b'a man a can a planal canada'
        python_stream = MockPythonStream(src_data)
        self._test(python_stream, src_data)


class Pkcs11LibTest(NativeResourceTest):
    def _lib_path(self):
        val = os.environ.get("AWS_TEST_PKCS11_LIB")
        if not val:
            raise unittest.SkipTest(f"test requires env var: AWS_TEST_PKCS11_LIB")
        return val

    def test_init(self):
        # sanity check that we can create/destroy
        lib_path = self._lib_path()
        pcks11_lib = Pkcs11Lib(file=lib_path, behavior=Pkcs11Lib.InitializeFinalizeBehavior.STRICT)

    def test_exceptions(self):
        # check that initialization errors bubble up as exceptions
        with self.assertRaises(Exception):
            pkcs11_lib = Pkcs11Lib(file='obviously-invalid-path.so')

        with self.assertRaises(Exception):
            with open(self._lib_path()) as literal_open_file:
                # a filepath str should passed, not a literal open file
                pkcs11_lib = Pkcs11Lib(file=literal_open_file)

    def test_strict_behavior(self):
        lib_path = self._lib_path()
        lib1 = Pkcs11Lib(file=lib_path, behavior=Pkcs11Lib.InitializeFinalizeBehavior.STRICT)
        # InitializeFinalizeBehavior.STRICT behavior should fail if the PKCS#11 lib is already loaded
        with self.assertRaises(Exception):
            lib2 = Pkcs11Lib(file=lib_path, behavior=Pkcs11Lib.InitializeFinalizeBehavior.STRICT)

    def test_omit_behavior(self):
        lib_path = self._lib_path()
        # InitializeFinalizeBehavior.OMIT should fail unless another instance of the PKCS#11 lib is already loaded
        with self.assertRaises(Exception):
            lib = Pkcs11Lib(file=lib_path, behavior=Pkcs11Lib.InitializeFinalizeBehavior.OMIT)

        # InitializeFinalizeBehavior.OMIT behavior should be fine when another
        # instance of the PKCS#11 lib is already loaded
        strict_lib = Pkcs11Lib(file=lib_path, behavior=Pkcs11Lib.InitializeFinalizeBehavior.STRICT)
        omit_lib = Pkcs11Lib(file=lib_path, behavior=Pkcs11Lib.InitializeFinalizeBehavior.OMIT)

    def test_is_tls_cipher_supported(self):
        self.assertEqual(True, TlsCipherPref.DEFAULT.is_supported())


if __name__ == '__main__':
    unittest.main()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import os

from configuration import available_ports
from common import (
    Ciphers,
    Curves,
    ProviderOptions,
    Protocols,
    KemGroups,
    Certificates,
    pq_enabled,
)
from fixtures import managed_process  # noqa: F401
from providers import Provider, S2N, OpenSSL, BoringSSL
from utils import invalid_test_parameters, get_parameter_name, to_bytes
from global_flags import get_flag, S2N_PROVIDER_VERSION

PQ_ENABLED_FLAG = "PQ key exchange enabled"

CIPHERS = [
    None,  # `None` will default to the appropriate `test_all` cipher preference in the S2N client provider
    Ciphers.KMS_TLS_1_0_2018_10,
    Ciphers.PQ_TLS_1_3_2023_06_01,
]

KEM_GROUPS = [
    KemGroups.X25519_KYBER512R3,
    KemGroups.P256_KYBER512R3,
    KemGroups.P384_KYBER768R3,
    KemGroups.P521_KYBER1024R3,
]

EXPECTED_RESULTS = {
    # The tuple keys have the form:
    # (client_{cipher, kem_group}, server_{cipher, kem_group}): {"cipher": {expected_cipher}, "kem_group": {expected_kem_group}}
    (Ciphers.PQ_TLS_1_0_2023_01, Ciphers.PQ_TLS_1_0_2023_01): {
        "cipher": "TLS_AES_256_GCM_SHA384",
        "kem_group": "_kyber-512-r3",
    },
    (KemGroups.P384_KYBER768R3, Ciphers.PQ_TLS_1_3_2023_06_01): {
        "cipher": "AES256_GCM_SHA384",
        "kem_group": "secp384r1_kyber-768-r3",
    },
    (KemGroups.P521_KYBER1024R3, Ciphers.PQ_TLS_1_3_2023_06_01): {
        "cipher": "AES256_GCM_SHA384",
        "kem_group": "secp521r1_kyber-1024-r3",
    },
    (Ciphers.PQ_TLS_1_3_2023_06_01, KemGroups.X25519Kyber768Draft00): {
        "cipher": "TLS_AES_256_GCM_SHA384",
        "kem_group": "X25519Kyber768Draft00",
    },
    (Ciphers.PQ_TLS_1_3_2023_06_01, KemGroups.SecP256r1Kyber768Draft00): {
        "cipher": "TLS_AES_256_GCM_SHA384",
        "kem_group": "SecP256r1Kyber768Draft00",
    },
    (Ciphers.PQ_TLS_1_3_2023_06_01, Ciphers.PQ_TLS_1_3_2023_06_01): {
        "cipher": "TLS_AES_256_GCM_SHA384",
        "kem_group": "SecP256r1Kyber768Draft00",
    },
    (Ciphers.PQ_TLS_1_3_2023_06_01, Ciphers.KMS_TLS_1_0_2018_10): {
        "cipher": "ECDHE-RSA-AES256-GCM-SHA384",
        "kem_group": None,
    },
    (Ciphers.KMS_TLS_1_0_2018_10, Ciphers.PQ_TLS_1_3_2023_06_01): {
        "cipher": "ECDHE-RSA-AES128-GCM-SHA256",
        "kem_group": None,
    },
    (Ciphers.KMS_TLS_1_0_2018_10, Ciphers.KMS_TLS_1_0_2018_10): {
        "cipher": "ECDHE-RSA-AES256-GCM-SHA384",
        "kem_group": None,
    },
}

"""
Similar to invalid_test_parameters(), this validates the test parameters for
both client and server. Returns True if the test case using these parameters
should be skipped.
"""


def invalid_pq_handshake_test_parameters(*args, **kwargs):
    client_cipher_kwargs = kwargs.copy()
    client_cipher_kwargs["cipher"] = kwargs["client_cipher"]

    server_cipher_kwargs = kwargs.copy()
    server_cipher_kwargs["cipher"] = kwargs["server_cipher"]

    # `or` is correct: invalid_test_parameters() returns True if the parameters are invalid;
    # we want to return True here if either of the sets of parameters are invalid.
    return invalid_test_parameters(
        *args, **client_cipher_kwargs
    ) or invalid_test_parameters(*args, **server_cipher_kwargs)


def get_oqs_openssl_override_env_vars():
    oqs_openssl_install_dir = os.environ["OQS_OPENSSL_1_1_1_INSTALL_DIR"]

    override_env_vars = dict()
    override_env_vars["PATH"] = oqs_openssl_install_dir + "/bin"
    override_env_vars["LD_LIBRARY_PATH"] = oqs_openssl_install_dir + "/lib"

    return override_env_vars


def assert_s2n_negotiation_parameters(s2n_results, expected_result):
    if expected_result is not None:
        assert (
            to_bytes(("Cipher negotiated: " + expected_result["cipher"]))
            in s2n_results.stdout
        )
        if expected_result["kem_group"]:
            # Purposefully leave off the "KEM Group: " prefix in order to perform partial matches
            # without specifying the curve.
            assert to_bytes(expected_result["kem_group"]) in s2n_results.stdout
            assert to_bytes(PQ_ENABLED_FLAG) in s2n_results.stdout
        if not expected_result["kem_group"]:
            assert to_bytes(PQ_ENABLED_FLAG) not in s2n_results.stdout
            assert to_bytes("Curve:") in s2n_results.stdout


def assert_awslc_negotiation_parameters(awslc_results, expected_result):
    assert expected_result is not None
    assert awslc_results.exit_code == 0
    assert to_bytes(("group: " + expected_result["kem_group"])) in awslc_results.stderr
    assert to_bytes(("Cipher: " + expected_result["cipher"])) in awslc_results.stderr


def test_nothing():
    """
    Sometimes the pq handshake test parameters in combination with the s2n libcrypto
    results in no test cases existing. In this case, pass a nothing test to avoid
    marking the entire codebuild run as failed.
    """
    assert True


@pytest.mark.uncollect_if(func=invalid_pq_handshake_test_parameters)
@pytest.mark.parametrize(
    "protocol", [Protocols.TLS12, Protocols.TLS13], ids=get_parameter_name
)
@pytest.mark.parametrize(
    "certificate", [Certificates.RSA_4096_SHA512], ids=get_parameter_name
)
@pytest.mark.parametrize("client_cipher", CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("server_cipher", CIPHERS, ids=get_parameter_name)
@pytest.mark.parametrize("provider", [S2N], ids=get_parameter_name)
@pytest.mark.parametrize("other_provider", [S2N], ids=get_parameter_name)
def test_s2nc_to_s2nd_pq_handshake(
    managed_process,  # noqa: F811
    protocol,
    certificate,
    client_cipher,
    server_cipher,
    provider,
    other_provider,
):
    port = next(available_ports)

    client_options = ProviderOptions(
        mode=Provider.ClientMode,
        port=port,
        insecure=True,
        cipher=client_cipher,
        protocol=protocol,
    )

    server_options = ProviderOptions(
        mode=Provider.ServerMode,
        port=port,
        protocol=protocol,
        cipher=server_cipher,
        cert=certificate.cert,
        key=certificate.key,
    )

    server = managed_process(S2N, server_options, timeout=5)
    client = managed_process(S2N, client_options, timeout=5)

    if pq_enabled():
        expected_result = EXPECTED_RESULTS.get((client_cipher, server_cipher), None)
    else:
        # If PQ is not enabled in s2n, we expect classic handshakes to be negotiated.
        # Leave the expected cipher blank, as there are multiple possibilities - the
        # important thing is that kem and kem_group are NONE.
        expected_result = {"cipher": "", "kem_group": None}

    # Client and server are both s2n; can make meaningful assertions about negotiation for both
    for results in client.get_results():
        results.assert_success()
        assert_s2n_negotiation_parameters(results, expected_result)

    for results in server.get_results():
        results.assert_success()
        assert_s2n_negotiation_parameters(results, expected_result)


@pytest.mark.parametrize(
    "s2n_client_policy", [Ciphers.PQ_TLS_1_3_2023_06_01], ids=get_parameter_name
)
@pytest.mark.parametrize(
    "awslc_server_group",
    [KemGroups.SecP256r1Kyber768Draft00, KemGroups.X25519Kyber768Draft00],
    ids=get_parameter_name,
)
def test_s2nc_to_awslc_pq_handshake(
    managed_process,  # noqa: F811
    s2n_client_policy,
    awslc_server_group,
):
    if not pq_enabled():
        pytest.skip("PQ not enabled")

    if "awslc" not in get_flag(S2N_PROVIDER_VERSION):
        pytest.skip(
            "s2n must be compiled with awslc libcrypto in order to test PQ TLS compatibility"
        )

    if "fips" in get_flag(S2N_PROVIDER_VERSION):
        pytest.skip(
            "No FIPS validated version of AWS-LC has support for negotiating Hybrid PQ TLS yet"
        )

    port = next(available_ports)

    s2nc_client_options = ProviderOptions(
        mode=Provider.ClientMode,
        port=port,
        insecure=True,
        cipher=s2n_client_policy,
        protocol=Protocols.TLS13,
    )

    awslc_server_options = ProviderOptions(
        mode=Provider.ServerMode,
        port=port,
        protocol=Protocols.TLS13,
        curve=Curves.from_name(awslc_server_group.oqs_name),
    )

    awslc_server = managed_process(BoringSSL, awslc_server_options, timeout=5)
    s2n_client = managed_process(S2N, s2nc_client_options, timeout=5)
    expected_result = EXPECTED_RESULTS.get(
        (s2n_client_policy, awslc_server_group), None
    )

    awslc_result = next(awslc_server.get_results())
    assert_awslc_negotiation_parameters(awslc_result, expected_result)

    s2nd_result = next(s2n_client.get_results())
    assert_s2n_negotiation_parameters(s2nd_result, expected_result)


@pytest.mark.parametrize(
    "s2n_server_policy", [Ciphers.PQ_TLS_1_3_2023_06_01], ids=get_parameter_name
)
@pytest.mark.parametrize(
    "awslc_client_group",
    [KemGroups.SecP256r1Kyber768Draft00, KemGroups.X25519Kyber768Draft00],
    ids=get_parameter_name,
)
def test_s2nd_to_awslc_pq_handshake(
    managed_process,  # noqa: F811
    s2n_server_policy,
    awslc_client_group,
):
    if not pq_enabled():
        pytest.skip("PQ not enabled")

    if "awslc" not in get_flag(S2N_PROVIDER_VERSION):
        pytest.skip(
            "s2n must be compiled with awslc libcrypto in order to test PQ TLS compatibility"
        )

    if "fips" in get_flag(S2N_PROVIDER_VERSION):
        pytest.skip(
            "No FIPS validated version of AWS-LC has support for negotiating Hybrid PQ TLS yet"
        )

    port = next(available_ports)

    s2nd_server_options = ProviderOptions(
        mode=Provider.ServerMode,
        port=port,
        insecure=True,
        cipher=s2n_server_policy,
        protocol=Protocols.TLS13,
    )

    awslc_client_options = ProviderOptions(
        mode=Provider.ClientMode,
        port=port,
        protocol=Protocols.TLS13,
        curve=Curves.from_name(awslc_client_group.oqs_name),
    )

    s2nd_server = managed_process(S2N, s2nd_server_options, timeout=5)
    awslc_client = managed_process(BoringSSL, awslc_client_options, timeout=5)
    expected_result = EXPECTED_RESULTS.get(
        (s2n_server_policy, awslc_client_group), None
    )

    awslc_result = next(awslc_client.get_results())
    assert_awslc_negotiation_parameters(awslc_result, expected_result)

    s2nd_result = next(s2nd_server.get_results())
    assert_s2n_negotiation_parameters(s2nd_result, expected_result)


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("protocol", [Protocols.TLS13], ids=get_parameter_name)
@pytest.mark.parametrize(
    "cipher", [Ciphers.PQ_TLS_1_3_2023_06_01], ids=get_parameter_name
)
@pytest.mark.parametrize("kem_group", KEM_GROUPS, ids=get_parameter_name)
def test_s2nc_to_oqs_openssl_pq_handshake(managed_process, protocol, cipher, kem_group):  # noqa: F811
    # If PQ is not enabled in s2n, there is no reason to test against oqs_openssl
    if not pq_enabled():
        return

    port = next(available_ports)

    client_options = ProviderOptions(
        mode=Provider.ClientMode,
        port=port,
        insecure=True,
        cipher=cipher,
        protocol=protocol,
    )

    server_options = ProviderOptions(
        mode=Provider.ServerMode,
        port=port,
        protocol=protocol,
        cert=Certificates.RSA_4096_SHA512.cert,
        key=Certificates.RSA_4096_SHA512.key,
        env_overrides=get_oqs_openssl_override_env_vars(),
        extra_flags=["-groups", kem_group.oqs_name],
    )

    server = managed_process(OpenSSL, server_options, timeout=5)
    client = managed_process(S2N, client_options, timeout=5)

    expected_result = EXPECTED_RESULTS.get((cipher, kem_group), None)

    for results in client.get_results():
        # Client is s2n; can make meaningful assertions about negotiation
        results.assert_success()
        assert_s2n_negotiation_parameters(results, expected_result)

    for results in server.get_results():
        # Server is OQS OpenSSL; just ensure the process exited successfully
        results.assert_success()


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize("protocol", [Protocols.TLS13], ids=get_parameter_name)
@pytest.mark.parametrize(
    "cipher", [Ciphers.PQ_TLS_1_3_2023_06_01], ids=get_parameter_name
)
@pytest.mark.parametrize("kem_group", KEM_GROUPS, ids=get_parameter_name)
def test_oqs_openssl_to_s2nd_pq_handshake(managed_process, protocol, cipher, kem_group):  # noqa: F811
    # If PQ is not enabled in s2n, there is no reason to test against oqs_openssl
    if not pq_enabled():
        return

    port = next(available_ports)

    client_options = ProviderOptions(
        mode=Provider.ClientMode,
        port=port,
        protocol=protocol,
        env_overrides=get_oqs_openssl_override_env_vars(),
        extra_flags=["-groups", kem_group.oqs_name],
    )

    server_options = ProviderOptions(
        mode=Provider.ServerMode,
        port=port,
        protocol=protocol,
        cipher=cipher,
        cert=Certificates.RSA_4096_SHA512.cert,
        key=Certificates.RSA_4096_SHA512.key,
    )

    server = managed_process(S2N, server_options, timeout=5)
    client = managed_process(OpenSSL, client_options, timeout=5)

    expected_result = EXPECTED_RESULTS.get((kem_group, cipher), None)

    for results in client.get_results():
        # Client is OQS OpenSSL; just ensure the process exited successfully
        results.assert_success()

    for results in server.get_results():
        # Server is s2n; can make meaningful assertions about negotiation
        results.assert_success()
        assert_s2n_negotiation_parameters(results, expected_result)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import copy
import os
from enum import Enum, auto

from configuration import available_ports
from common import Ciphers, ProviderOptions, Protocols, random_str
from fixtures import managed_process  # noqa: F401
from providers import Provider, S2N
from utils import invalid_test_parameters, get_parameter_name, to_bytes

SERVER_STATE_FILE = "server_state"
CLIENT_STATE_FILE = "client_state"

SERVER_DATA = "Some random data from the server:" + random_str(10)
CLIENT_DATA = "Some random data from the client:" + random_str(10)

S2N_TEST_POLICIES = {
    Protocols.TLS12.value: Ciphers.SECURITY_POLICY_DEFAULT_TLS12,
    Protocols.TLS13.value: Ciphers.SECURITY_POLICY_DEFAULT_TLS13,
}


class MainlineRole(Enum):
    Serialize = auto()
    Deserialize = auto()


class Mode(Enum):
    Server = auto()
    Client = auto()


"""
This test file checks that a serialized connection can be deserialized by an older version of
s2n-tls and vice versa. This ensures that any future changes we make to the handshake are backwards-compatible
with an older version of s2n-tls.

This feature requires an uninterrupted TCP connection with the peer in-between serialization and
deserialization. Our integration test setup can't easily provide that while also using two different
s2n-tls versions. To get around that we do a hack and serialize/deserialize both peers in the TLS connection.
This prevents one peer from receiving a TCP FIN message and shutting the connection down early.
"""


@pytest.mark.uncollect_if(func=invalid_test_parameters)
@pytest.mark.parametrize(
    "protocol", [Protocols.TLS13, Protocols.TLS12], ids=get_parameter_name
)
@pytest.mark.parametrize(
    "mainline_role",
    [MainlineRole.Serialize, MainlineRole.Deserialize],
    ids=get_parameter_name,
)
@pytest.mark.parametrize(
    "version_change", [Mode.Server, Mode.Client], ids=get_parameter_name
)
def test_server_serialization_backwards_compat(
    managed_process,  # noqa: F811
    tmp_path,
    protocol,
    mainline_role,
    version_change,
):
    server_state_file = str(tmp_path / SERVER_STATE_FILE)
    client_state_file = str(tmp_path / CLIENT_STATE_FILE)
    assert not os.path.exists(server_state_file)
    assert not os.path.exists(client_state_file)

    options = ProviderOptions(
        port=next(available_ports),
        cipher=S2N_TEST_POLICIES[protocol.value],
        insecure=True,
        echo=False,
    )

    client_options = copy.copy(options)
    client_options.mode = Provider.ClientMode
    client_options.extra_flags = ["--serialize-out", client_state_file]

    server_options = copy.copy(options)
    server_options.mode = Provider.ServerMode
    server_options.extra_flags = ["--serialize-out", server_state_file]

    if mainline_role is MainlineRole.Serialize:
        if version_change == Mode.Server:
            server_options.use_mainline_version = True
        else:
            client_options.use_mainline_version = True

    server = managed_process(S2N, server_options, send_marker=S2N.get_send_marker())
    client = managed_process(S2N, client_options, send_marker=S2N.get_send_marker())

    for results in client.get_results():
        results.assert_success()
        assert (
            to_bytes("Actual protocol version: {}".format(protocol.value))
            in results.stdout
        )

    for results in server.get_results():
        results.assert_success()
        assert (
            to_bytes("Actual protocol version: {}".format(protocol.value))
            in results.stdout
        )

    assert os.path.exists(server_state_file)
    assert os.path.exists(client_state_file)

    client_options.extra_flags = ["--deserialize-in", client_state_file]
    server_options.extra_flags = ["--deserialize-in", server_state_file]
    if mainline_role is MainlineRole.Deserialize:
        if version_change == Mode.Server:
            server_options.use_mainline_version = True
        else:
            client_options.use_mainline_version = True

    server_options.echo = True
    server_options.data_to_send = SERVER_DATA.encode()

    client_options.echo = True
    client_options.data_to_send = CLIENT_DATA.encode()

    server = managed_process(S2N, server_options, send_marker=CLIENT_DATA)
    client = managed_process(
        S2N,
        client_options,
        send_marker="Connected to localhost",
        close_marker=SERVER_DATA,
    )

    for results in server.get_results():
        results.assert_success()
        # No protocol version printout since deserialization means skipping the handshake
        assert to_bytes("Actual protocol version:") not in results.stdout
        assert CLIENT_DATA.encode() in results.stdout

    for results in client.get_results():
        results.assert_success()
        assert to_bytes("Actual protocol version:") not in results.stdout
        assert SERVER_DATA.encode() in results.stdout



import os
import shlex
import subprocess
import sys
from typing import Sequence, Union


def run(args: Union[str, Sequence[str]]):
    """
    Run a program.

    args may be a string, or list of argument strings.

    If the program is "python3", this is replaced with the
    full path to the current python executable.
    """

    # convert string to list
    # so that we don't need to pass shell=True to subprocess.run()
    # because turning on shell can mess things up (i.e. clang-format hangs forever for some reason)
    if isinstance(args, str):
        args = shlex.split(args)

    # ensure proper python executable is used
    if args[0] == 'python3':
        args[0] = sys.executable

    # run
    print(f'+ {subprocess.list2cmdline(args)}')
    result = subprocess.run(args)
    if result.returncode != 0:
        sys.exit('FAILED')


def chdir_project_root():
    root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    os.chdir(root)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0.

import socket
import ssl

context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
context.minimum_version = ssl.TLSVersion.TLSv1_3
context.maximum_version = ssl.TLSVersion.TLSv1_3
context.load_cert_chain('../resources/tls13_server.pem.crt', '../resources/tls13_server.key')
context.load_verify_locations('../resources/tls13_device_root_ca.pem.crt')
context.verify_mode = ssl.CERT_REQUIRED

with socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0) as sock:
    sock.bind(('127.0.0.1', 59443))
    sock.listen(1)
    with context.wrap_socket(sock, server_side=True) as ssock:
        while True:
            try:
                conn, addr = ssock.accept()
                print("accepted new connection: {}".format(addr))
            except Exception as e:
                print("accept failed: {}".format(e))



#!/usr/bin/env python3
import glob
import os
import shutil
import utils

# apply these patterns without recursing through subfolders
NONRECURSIVE_PATTERNS = [
    'build/',
    '_awscrt*.*',  # compiled _awscrt shared lib
    '*.egg-info/',
    'dist/',
    'wheelhouse/',
    'docsrc/build/',
]

# recurse through subfolders and apply these patterns
RECURSIVE_PATTERNS = [
    '*.pyc',
    '__pycache__'
]

# approved list of folders
# because we don't want to clean the virtual environment folder
APPROVED_RECURSIVE_FOLDERS = [
    'awscrt',
    'scripts',
    'test',
    '.builder',
]

utils.chdir_project_root()

utils.run('python3 -m pip uninstall -y awscrt')

paths = []
for pattern in NONRECURSIVE_PATTERNS:
    paths.extend(glob.glob(pattern))

for pattern in RECURSIVE_PATTERNS:
    for folder in APPROVED_RECURSIVE_FOLDERS:
        paths.extend(glob.glob(f'{folder}/**/{pattern}', recursive=True))

for path in paths:
    print(f'delete: {path}')
    if os.path.isfile(path):
        os.remove(path)
    else:
        shutil.rmtree(path)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0.
from awscrt.mqtt_request_response import StreamingOperationOptions
from test import test_retry_wrapper, NativeResourceTest
from awscrt import io, mqtt5, mqtt_request_response, mqtt

from concurrent.futures import Future
import os
import time
import unittest
import uuid

TIMEOUT = 30.0


def create_client_id():
    return f"aws-crt-python-unit-test-{uuid.uuid4()}"


def _get_env_variable(env_name):
    env_data = os.environ.get(env_name)
    if not env_data:
        raise unittest.SkipTest(f"test requires env var: {env_name}")
    return env_data


class MqttRequestResponse5TestCallbacks():
    def __init__(self):
        self.future_connection_success = Future()
        self.future_stopped = Future()

    def ws_handshake_transform(self, transform_args):
        transform_args.set_done()

    def on_publish_received(self, publish_received_data: mqtt5.PublishReceivedData):
        pass

    def on_lifecycle_stopped(self, lifecycle_stopped: mqtt5.LifecycleStoppedData):
        if self.future_stopped:
            self.future_stopped.set_result(None)

    def on_lifecycle_attempting_connect(self, lifecycle_attempting_connect: mqtt5.LifecycleAttemptingConnectData):
        pass

    def on_lifecycle_connection_success(self, lifecycle_connection_success: mqtt5.LifecycleConnectSuccessData):
        if self.future_connection_success:
            if self.future_connection_success.done():
                pass
            else:
                self.future_connection_success.set_result(lifecycle_connection_success)

    def on_lifecycle_connection_failure(self, lifecycle_connection_failure: mqtt5.LifecycleConnectFailureData):
        if self.future_connection_success:
            if self.future_connection_success.done():
                pass
            else:
                self.future_connection_success.set_exception(lifecycle_connection_failure.exception)

    def on_lifecycle_disconnection(self, lifecycle_disconnect_data: mqtt5.LifecycleDisconnectData):
        pass


def _empty_response_paths(options):
    options.response_paths = []


def _invalidate_response_path_topic(options):
    options.response_paths[0].topic = "a/#/b"


def _none_response_path_topic(options):
    options.response_paths[0].topic = None


def _missing_response_path_topic(options):
    del options.response_paths[0].topic


def _type_mismatch_response_path_topic(options):
    options.response_paths[0].topic = 57.3


def _type_mismatch_response_path_correlation_token_json_path(options):
    options.response_paths[0].correlation_token_json_path = []


def _type_mismatch_response_paths(options):
    options.response_paths = "hello"


def _invalidate_subscription_topic_filter(options):
    options.subscription_topic_filters[0] = "a/#/c"


def _type_mismatch_subscription_topic_filter(options):
    options.subscription_topic_filters[0] = ["thirty", 30]


def _type_mismatch_subscriptions(options):
    options.subscription_topic_filters = 50


def _empty_subscription_topic_filters(options):
    options.subscription_topic_filters = []


def _none_publish_topic(options):
    options.publish_topic = None


def _bad_publish_topic(options):
    options.publish_topic = "#/b/c"


def _type_mismatch_publish_topic(options):
    options.publish_topic = [["oof"]]


def _type_mismatch_correlation_token(options):
    options.correlation_token = [-1]


def _subscription_topic_filter_none(options):
    options.subscription_topic_filter = None


def _type_mismatch_stream_subscription_topic_filter(options):
    options.subscription_topic_filter = 6


def _type_mismatch_subscription_status_listener(options):
    options.subscription_status_listener = "string"


def _type_mismatch_incoming_publish_listener(options):
    options.incoming_publish_listener = "string"


class MqttRequestResponseClientTest(NativeResourceTest):

    def _create_client5(self):

        input_host_name = _get_env_variable("AWS_TEST_MQTT5_IOT_CORE_HOST")
        input_cert = _get_env_variable("AWS_TEST_MQTT5_IOT_CORE_RSA_CERT")
        input_key = _get_env_variable("AWS_TEST_MQTT5_IOT_CORE_RSA_KEY")

        client_options = mqtt5.ClientOptions(
            host_name=input_host_name,
            port=8883
        )

        tls_ctx_options = io.TlsContextOptions.create_client_with_mtls_from_path(
            input_cert,
            input_key
        )
        client_options.tls_ctx = io.ClientTlsContext(tls_ctx_options)

        client_options.connect_options = mqtt5.ConnectPacket()
        client_options.connect_options.client_id = create_client_id()

        callbacks = MqttRequestResponse5TestCallbacks()
        client_options.on_lifecycle_event_stopped_fn = callbacks.on_lifecycle_stopped
        client_options.on_lifecycle_event_connection_success_fn = callbacks.on_lifecycle_connection_success
        client_options.on_lifecycle_event_connection_failure_fn = callbacks.on_lifecycle_connection_failure
        client_options.on_lifecycle_event_stopped_fn = callbacks.on_lifecycle_stopped

        protocol_client = mqtt5.Client(client_options)
        protocol_client.start()

        callbacks.future_connection_success.result(TIMEOUT)

        return protocol_client, callbacks

    def _shutdown5(self, protocol_client, callbacks):

        protocol_client.stop()
        callbacks.future_stopped.result(TIMEOUT)

    def _create_client311(self):

        input_host_name = _get_env_variable("AWS_TEST_MQTT5_IOT_CORE_HOST")
        input_cert = _get_env_variable("AWS_TEST_MQTT5_IOT_CORE_RSA_CERT")
        input_key = _get_env_variable("AWS_TEST_MQTT5_IOT_CORE_RSA_KEY")

        tls_ctx_options = io.TlsContextOptions.create_client_with_mtls_from_path(
            input_cert,
            input_key
        )
        tls_ctx = io.ClientTlsContext(tls_ctx_options)

        client = mqtt.Client(None, tls_ctx)

        protocol_client = mqtt.Connection(
            client=client,
            client_id=create_client_id(),
            host_name=input_host_name,
            port=8883,
            ping_timeout_ms=10000,
            keep_alive_secs=30
        )
        protocol_client.connect().result(TIMEOUT)

        return protocol_client

    def _shutdown311(self, protocol_client):
        protocol_client.disconnect().result(TIMEOUT)

    def _create_rr_client(
            self,
            protocol_client,
            max_request_response_subscriptions,
            max_streaming_subscriptions,
            operation_timeout_seconds):
        rr_client_options = mqtt_request_response.ClientOptions(
            max_request_response_subscriptions, max_streaming_subscriptions)
        rr_client_options.operation_timeout_in_seconds = operation_timeout_seconds

        rr_client = mqtt_request_response.Client(protocol_client, rr_client_options)

        return rr_client

    def _do_mqtt5_test(self, test_callable):
        (protocol_client, callbacks) = self._create_client5()

        test_callable(protocol_client)

        self._shutdown5(protocol_client, callbacks)

    def _do_mqtt311_test(self, test_callable):
        protocol_client = self._create_client311()

        test_callable(protocol_client)

        self._shutdown311(protocol_client)

    def _create_get_shadow_request(self, thing_name, use_correlation_token):
        topic_prefix = f"$aws/things/{thing_name}/shadow/get"

        request_options = mqtt_request_response.RequestOptions(
            subscription_topic_filters=[f"{topic_prefix}/+"],
            response_paths=[
                mqtt_request_response.ResponsePath(topic=f"{topic_prefix}/accepted"),
                mqtt_request_response.ResponsePath(topic=f"{topic_prefix}/rejected")
            ],
            publish_topic=topic_prefix,
            payload="{}".encode()
        )

        if use_correlation_token:
            correlation_token = f"{uuid.uuid4()}"
            request_options.response_paths[0].correlation_token_json_path = "clientToken"
            request_options.response_paths[1].correlation_token_json_path = "clientToken"
            request_options.payload = f'{{"clientToken":"{correlation_token}"}}'.encode()
            request_options.correlation_token = correlation_token

        return request_options

    def _do_get_shadow_success_no_such_shadow(self, rr_client, thing_name, use_correlation_token):
        request_options = self._create_get_shadow_request(thing_name, use_correlation_token)

        request_future = rr_client.make_request(request_options)
        response = request_future.result()

        self.assertEqual(request_options.response_paths[1].topic, response.topic,
                         "Expected response to come in on rejected topic")
        payload = str(response.payload)

        self.assertIn("No shadow exists with name", payload)

    def _do_get_shadow_success_no_such_shadow_test(self, protocol_client, use_correlation_token):
        rr_client = self._create_rr_client(protocol_client, 2, 2, 30)

        thing_name = f"tn-{uuid.uuid4()}"
        self._do_get_shadow_success_no_such_shadow(rr_client, thing_name, use_correlation_token)

    def _do_get_shadow_success(self, rr_client, thing_name, use_correlation_token):
        request_options = self._create_get_shadow_request(thing_name, use_correlation_token)

        request_future = rr_client.make_request(request_options)
        response = request_future.result()

        self.assertEqual(request_options.response_paths[0].topic, response.topic,
                         "Expected response to come in on accepted topic")

        payload = str(response.payload)
        self.assertIn("magic", payload)

    def _do_update_shadow_success(self, rr_client, thing_name, use_correlation_token):
        topic_prefix = f"$aws/things/{thing_name}/shadow/update"

        request_options = mqtt_request_response.RequestOptions(
            subscription_topic_filters=[f"{topic_prefix}/accepted", f"{topic_prefix}/rejected"],
            response_paths=[
                mqtt_request_response.ResponsePath(topic=f"{topic_prefix}/accepted"),
                mqtt_request_response.ResponsePath(topic=f"{topic_prefix}/rejected")
            ],
            publish_topic=topic_prefix,
            payload="".encode()
        )

        desired_state = f'{{"magic":"value"}}'

        if use_correlation_token:
            correlation_token = f"{uuid.uuid4()}"
            payload = f'{{"clientToken": "{correlation_token}","state": {{"desired": {desired_state}}}}}'
            request_options.response_paths[0].correlation_token_json_path = "clientToken"
            request_options.response_paths[1].correlation_token_json_path = "clientToken"
            request_options.payload = payload.encode()
            request_options.correlation_token = correlation_token
        else:
            request_options.payload = f'{{"state":{{"desired":{desired_state}}}}}'.encode()

        request_future = rr_client.make_request(request_options)
        response = request_future.result()

        self.assertEqual(request_options.response_paths[0].topic, response.topic,
                         "Expected response to come in on accepted topic")

        payload = str(response.payload)
        self.assertIn("magic", payload)

    def _do_delete_shadow_success(self, rr_client, thing_name, use_correlation_token):
        topic_prefix = f"$aws/things/{thing_name}/shadow/delete"

        request_options = mqtt_request_response.RequestOptions(
            subscription_topic_filters=[f"{topic_prefix}/+"],
            response_paths=[
                mqtt_request_response.ResponsePath(topic=f"{topic_prefix}/accepted"),
                mqtt_request_response.ResponsePath(topic=f"{topic_prefix}/rejected")
            ],
            publish_topic=topic_prefix,
            payload="{}".encode()
        )

        if use_correlation_token:
            correlation_token = f"{uuid.uuid4()}"
            request_options.response_paths[0].correlation_token_json_path = "clientToken"
            request_options.response_paths[1].correlation_token_json_path = "clientToken"
            request_options.payload = f'{{"clientToken":"{correlation_token}"}}'.encode()
            request_options.correlation_token = correlation_token

        request_future = rr_client.make_request(request_options)
        response = request_future.result()

        self.assertEqual(request_options.response_paths[0].topic, response.topic,
                         "Expected response to come in on accepted topic")

        payload = str(response.payload)
        self.assertIn("version", payload)

    def _do_update_delete_shadow_success_test(self, protocol_client, use_correlation_token):
        rr_client = self._create_rr_client(protocol_client, 2, 2, 30)

        # get should return non-existence
        thing_name = f"tn-{uuid.uuid4()}"

        try:
            self._do_get_shadow_success_no_such_shadow(rr_client, thing_name, use_correlation_token)

            # update shadow to create
            self._do_update_shadow_success(rr_client, thing_name, use_correlation_token)

            # eventual consistency worries
            time.sleep(2)

            # get should now return the shadow state
            self._do_get_shadow_success(rr_client, thing_name, use_correlation_token)
        finally:
            # delete shadow
            self._do_delete_shadow_success(rr_client, thing_name, use_correlation_token)

    def _do_stream_success_test(self, protocol_client):
        rr_client = self._create_rr_client(protocol_client, 2, 2, 30)

        thing_name = f"tn-{uuid.uuid4()}"
        topic = f"not/a/valid/topic/{thing_name}"
        payload = b'hello'

        subscribed_future = Future()
        publish_received_future = Future()

        def on_subscription_status_changed(event):
            if event.type == mqtt_request_response.SubscriptionStatusEventType.SUBSCRIPTION_ESTABLISHED:
                subscribed_future.set_result(True)

        def on_incoming_publish(event):
            if event.topic == topic and event.payload == payload:
                publish_received_future.set_result(True)

        stream_options = StreamingOperationOptions(topic, on_subscription_status_changed, on_incoming_publish)
        stream = rr_client.create_stream(stream_options)

        stream.open()

        assert subscribed_future.result(30)

        if isinstance(protocol_client, mqtt5.Client):
            publish = mqtt5.PublishPacket(
                payload=payload,
                qos=mqtt5.QoS.AT_LEAST_ONCE,
                topic=topic
            )
            protocol_client.publish(publish)
        else:
            protocol_client.publish(topic, payload, mqtt.QoS.AT_LEAST_ONCE)

        assert publish_received_future.result(30)

    def _do_get_shadow_failure_test(self, protocol_client, options_transform):
        rr_client = self._create_rr_client(protocol_client, 2, 2, 30)

        thing_name = f"tn-{uuid.uuid4()}"
        request_options = self._create_get_shadow_request(thing_name, True)
        options_transform(request_options)

        self.assertRaises(Exception, lambda: rr_client.make_request(request_options))

    def _do_get_shadow_future_failure_test(self, protocol_client, options_transform):
        rr_client = self._create_rr_client(protocol_client, 2, 2, 30)

        thing_name = f"tn-{uuid.uuid4()}"
        request_options = self._create_get_shadow_request(thing_name, True)
        options_transform(request_options)

        response_future = rr_client.make_request(request_options)

        self.assertRaises(Exception, lambda: response_future.result())

    def _do_create_stream_failure_test(self, protocol_client, options_transform):
        rr_client = self._create_rr_client(protocol_client, 2, 2, 30)

        thing_name = f"tn-{uuid.uuid4()}"
        stream_options = mqtt_request_response.StreamingOperationOptions(f"not/a/real/thing/{thing_name}")
        options_transform(stream_options)

        self.assertRaises(Exception, lambda: rr_client.create_stream(stream_options))

    # ==============================================================
    #             CREATION SUCCESS TEST CASES
    # ==============================================================

    def _test_client_creation_success5(self):
        self._do_mqtt5_test(lambda protocol_client: self._create_rr_client(protocol_client, 2, 2, 30))

    def test_client_creation_success5(self):
        test_retry_wrapper(self._test_client_creation_success5)

    def _test_client_creation_success311(self):
        self._do_mqtt311_test(lambda protocol_client: self._create_rr_client(protocol_client, 2, 2, 30))

    def test_client_creation_success311(self):
        test_retry_wrapper(self._test_client_creation_success311)

    # ==============================================================
    #             CREATION FAILURE TEST CASES
    # ==============================================================

    def test_client_creation_failure_no_protocol_client(self):
        self.assertRaises(Exception, self._create_rr_client, None, 2, 2, 30)

    def test_client_creation_failure_zero_request_response_subscriptions5(self):
        self._do_mqtt5_test(lambda protocol_client: self.assertRaises(
            Exception, self._create_rr_client, protocol_client, 0, 2, 30))

    def test_client_creation_failure_zero_request_response_subscriptions311(self):
        self._do_mqtt311_test(lambda protocol_client: self.assertRaises(
            Exception, self._create_rr_client, protocol_client, 0, 2, 30))

    def test_client_creation_failure_negative_request_response_subscriptions5(self):
        self._do_mqtt5_test(lambda protocol_client: self.assertRaises(
            Exception, self._create_rr_client, protocol_client, -2, 2, 30))

    def test_client_creation_failure_negative_request_response_subscriptions311(self):
        self._do_mqtt311_test(lambda protocol_client: self.assertRaises(
            Exception, self._create_rr_client, protocol_client, -2, 2, 30))

    def test_client_creation_failure_no_request_response_subscriptions5(self):
        self._do_mqtt5_test(lambda protocol_client: self.assertRaises(
            Exception, self._create_rr_client, protocol_client, None, 2, 30))

    def test_client_creation_failure_no_request_response_subscriptions311(self):
        self._do_mqtt311_test(lambda protocol_client: self.assertRaises(
            Exception, self._create_rr_client, protocol_client, None, 2, 30))

    def test_client_creation_failure_request_response_subscriptions_type_mismatch5(self):
        self._do_mqtt5_test(
            lambda protocol_client: self.assertRaises(
                Exception,
                self._create_rr_client,
                protocol_client,
                "None",
                2,
                30))

    def test_client_creation_failure_request_response_subscriptions_type_mismatch311(self):
        self._do_mqtt311_test(lambda protocol_client: self.assertRaises(
            Exception, self._create_rr_client, protocol_client, "None", 2, 30))

    def test_client_creation_failure_negative_streaming_subscriptions5(self):
        self._do_mqtt5_test(lambda protocol_client: self.assertRaises(
            Exception, self._create_rr_client, protocol_client, 2, -2, 30))

    def test_client_creation_failure_negative_streaming_subscriptions311(self):
        self._do_mqtt311_test(lambda protocol_client: self.assertRaises(
            Exception, self._create_rr_client, protocol_client, 2, -2, 30))

    def test_client_creation_failure_no_streaming_subscriptions5(self):
        self._do_mqtt5_test(lambda protocol_client: self.assertRaises(
            Exception, self._create_rr_client, protocol_client, 2, None, 30))

    def test_client_creation_failure_no_streaming_subscriptions311(self):
        self._do_mqtt311_test(lambda protocol_client: self.assertRaises(
            Exception, self._create_rr_client, protocol_client, 2, None, 30))

    def test_client_creation_failure_streaming_subscriptions_type_mismatch5(self):
        self._do_mqtt5_test(lambda protocol_client: self.assertRaises(
            Exception, self._create_rr_client, protocol_client, 2, [], 30))

    def test_client_creation_failure_streaming_subscriptions_type_mismatch311(self):
        self._do_mqtt311_test(lambda protocol_client: self.assertRaises(
            Exception, self._create_rr_client, protocol_client, 2, {}, 30))

    def test_client_creation_failure_negative_operation_timeout5(self):
        self._do_mqtt5_test(lambda protocol_client: self.assertRaises(
            Exception, self._create_rr_client, protocol_client, 2, 2, -30))

    def test_client_creation_failure_negative_operation_timeout311(self):
        self._do_mqtt311_test(lambda protocol_client: self.assertRaises(
            Exception, self._create_rr_client, protocol_client, 2, 2, -30))

    def test_client_creation_failure_no_operation_timeout5(self):
        self._do_mqtt5_test(lambda protocol_client: self.assertRaises(
            Exception, self._create_rr_client, protocol_client, 2, 2, None))

    def test_client_creation_failure_no_operation_timeout311(self):
        self._do_mqtt311_test(lambda protocol_client: self.assertRaises(
            Exception, self._create_rr_client, protocol_client, 2, 2, None))

    def test_client_creation_failure_operation_timeout_invalid5(self):
        self._do_mqtt5_test(lambda protocol_client: self.assertRaises(
            Exception, self._create_rr_client, protocol_client, 2, 2, ["uffdah"]))

    def test_client_creation_failure_operation_timeout_invalid311(self):
        self._do_mqtt311_test(
            lambda protocol_client: self.assertRaises(
                Exception,
                self._create_rr_client,
                protocol_client,
                2,
                2,
                777777777777777777777777777777777777))

    # ==============================================================
    #             make_request SUCCESS TEST CASES
    # ==============================================================

    def _test_get_shadow_success_no_such_shadow5(self):
        self._do_mqtt5_test(
            lambda protocol_client: self._do_get_shadow_success_no_such_shadow_test(
                protocol_client, True))

    def test_get_shadow_success_no_such_shadow5(self):
        test_retry_wrapper(self._test_get_shadow_success_no_such_shadow5)

    def _test_get_shadow_success_no_such_shadow311(self):
        self._do_mqtt311_test(
            lambda protocol_client: self._do_get_shadow_success_no_such_shadow_test(
                protocol_client, True))

    def test_get_shadow_success_no_such_shadow311(self):
        test_retry_wrapper(self._test_get_shadow_success_no_such_shadow311)

    def _test_get_shadow_success_no_such_shadow_no_correlation_token5(self):
        self._do_mqtt5_test(
            lambda protocol_client: self._do_get_shadow_success_no_such_shadow_test(
                protocol_client, False))

    def test_get_shadow_success_no_such_shadow_no_correlation_token5(self):
        test_retry_wrapper(self._test_get_shadow_success_no_such_shadow_no_correlation_token5)

    def _test_get_shadow_success_no_such_shadow_no_correlation_token311(self):
        self._do_mqtt311_test(
            lambda protocol_client: self._do_get_shadow_success_no_such_shadow_test(
                protocol_client, False))

    def test_get_shadow_success_no_such_shadow_no_correlation_token311(self):
        test_retry_wrapper(self._test_get_shadow_success_no_such_shadow_no_correlation_token311)

    def _test_update_delete_shadow_success5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_update_delete_shadow_success_test(protocol_client, True))

    def test_update_delete_shadow_success5(self):
        test_retry_wrapper(self._test_update_delete_shadow_success5)

    def _test_update_delete_shadow_success311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_update_delete_shadow_success_test(protocol_client, True))

    def test_update_delete_shadow_success311(self):
        test_retry_wrapper(self._test_update_delete_shadow_success311)

    def _test_update_delete_shadow_success_no_correlation_token5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_update_delete_shadow_success_test(protocol_client, False))

    def test_update_delete_shadow_success_no_correlation_token5(self):
        test_retry_wrapper(self._test_update_delete_shadow_success_no_correlation_token5)

    def _test_update_delete_shadow_success_no_correlation_token311(self):
        self._do_mqtt311_test(
            lambda protocol_client: self._do_update_delete_shadow_success_test(
                protocol_client, False))

    def test_update_delete_shadow_success_no_correlation_token311(self):
        test_retry_wrapper(self._test_update_delete_shadow_success_no_correlation_token311)

    # ==============================================================
    #             make_request FAILURE TEST CASES
    # ==============================================================

    def _test_get_shadow_failure_no_response_paths5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _empty_response_paths(options)))

    def test_get_shadow_failure_no_response_paths5(self):
        test_retry_wrapper(self._test_get_shadow_failure_no_response_paths5)

    def _test_get_shadow_failure_no_response_paths311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _empty_response_paths(options)))

    def test_get_shadow_failure_no_response_paths311(self):
        test_retry_wrapper(self._test_get_shadow_failure_no_response_paths311)

    def _test_get_shadow_failure_invalid_response_path_topic5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_get_shadow_future_failure_test(
            protocol_client, lambda options: _invalidate_response_path_topic(options)))

    def test_get_shadow_failure_invalid_response_path_topic5(self):
        test_retry_wrapper(self._test_get_shadow_failure_invalid_response_path_topic5)

    def _test_get_shadow_failure_invalid_response_path_topic311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_get_shadow_future_failure_test(
            protocol_client, lambda options: _invalidate_response_path_topic(options)))

    def test_get_shadow_failure_invalid_response_path_topic311(self):
        test_retry_wrapper(self._test_get_shadow_failure_invalid_response_path_topic311)

    def _test_get_shadow_failure_none_response_path_topic5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _none_response_path_topic(options)))

    def test_get_shadow_failure_none_response_path_topic5(self):
        test_retry_wrapper(self._test_get_shadow_failure_none_response_path_topic5)

    def _test_get_shadow_failure_none_response_path_topic311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _none_response_path_topic(options)))

    def test_get_shadow_failure_none_response_path_topic311(self):
        test_retry_wrapper(self._test_get_shadow_failure_none_response_path_topic311)

    def _test_get_shadow_failure_missing_response_path_topic5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _missing_response_path_topic(options)))

    def test_get_shadow_failure_missing_response_path_topic5(self):
        test_retry_wrapper(self._test_get_shadow_failure_missing_response_path_topic5)

    def _test_get_shadow_failure_missing_response_path_topic311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _missing_response_path_topic(options)))

    def test_get_shadow_failure_missing_response_path_topic311(self):
        test_retry_wrapper(self._test_get_shadow_failure_missing_response_path_topic311)

    def _test_get_shadow_failure_response_path_topic_type_mismatch5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _type_mismatch_response_path_topic(options)))

    def test_get_shadow_failure_response_path_topic_type_mismatch5(self):
        test_retry_wrapper(self._test_get_shadow_failure_response_path_topic_type_mismatch5)

    def _test_get_shadow_failure_response_path_topic_type_mismatch311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _type_mismatch_response_path_topic(options)))

    def test_get_shadow_failure_response_path_topic_type_mismatch311(self):
        test_retry_wrapper(self._test_get_shadow_failure_response_path_topic_type_mismatch311)

    def _test_get_shadow_failure_response_path_correlation_token_json_path_type_mismatch5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _type_mismatch_response_path_correlation_token_json_path(options)))

    def test_get_shadow_failure_response_path_correlation_token_json_path_type_mismatch5(self):
        test_retry_wrapper(self._test_get_shadow_failure_response_path_correlation_token_json_path_type_mismatch5)

    def _test_get_shadow_failure_response_path_correlation_token_json_path_type_mismatch311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _type_mismatch_response_path_correlation_token_json_path(options)))

    def test_get_shadow_failure_response_path_correlation_token_json_path_type_mismatch311(self):
        test_retry_wrapper(self._test_get_shadow_failure_response_path_correlation_token_json_path_type_mismatch311)

    def _test_get_shadow_failure_response_paths_type_mismatch5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _type_mismatch_response_paths(options)))

    def test_get_shadow_failure_response_paths_type_mismatch5(self):
        test_retry_wrapper(self._test_get_shadow_failure_response_paths_type_mismatch5)

    def _test_get_shadow_failure_response_paths_type_mismatch311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _type_mismatch_response_paths(options)))

    def test_get_shadow_failure_response_paths_type_mismatch311(self):
        test_retry_wrapper(self._test_get_shadow_failure_response_paths_type_mismatch311)

    def _test_get_shadow_failure_invalid_subscription_topic5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_get_shadow_future_failure_test(
            protocol_client, lambda options: _invalidate_subscription_topic_filter(options)))

    def test_get_shadow_failure_invalid_subscription_topic5(self):
        test_retry_wrapper(self._test_get_shadow_failure_invalid_subscription_topic5)

    def _test_get_shadow_failure_invalid_subscription_topic311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_get_shadow_future_failure_test(
            protocol_client, lambda options: _invalidate_subscription_topic_filter(options)))

    def test_get_shadow_failure_invalid_subscription_topic311(self):
        test_retry_wrapper(self._test_get_shadow_failure_invalid_subscription_topic311)

    def _test_get_shadow_failure_subscription_topic_type_mismatch5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _type_mismatch_subscription_topic_filter(options)))

    def test_get_shadow_failure_subscription_topic_type_mismatch5(self):
        test_retry_wrapper(self._test_get_shadow_failure_subscription_topic_type_mismatch5)

    def _test_get_shadow_failure_subscription_topic_type_mismatch311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _type_mismatch_subscription_topic_filter(options)))

    def test_get_shadow_failure_subscription_topic_type_mismatch311(self):
        test_retry_wrapper(self._test_get_shadow_failure_subscription_topic_type_mismatch311)

    def _test_get_shadow_failure_subscriptions_type_mismatch5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _type_mismatch_subscriptions(options)))

    def test_get_shadow_failure_subscriptions_type_mismatch5(self):
        test_retry_wrapper(self._test_get_shadow_failure_subscriptions_type_mismatch5)

    def _test_get_shadow_failure_subscriptions_type_mismatch311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _type_mismatch_subscriptions(options)))

    def test_get_shadow_failure_subscriptions_type_mismatch311(self):
        test_retry_wrapper(self._test_get_shadow_failure_subscriptions_type_mismatch311)

    def _test_get_shadow_failure_empty_subscriptions5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _empty_subscription_topic_filters(options)))

    def test_get_shadow_failure_empty_subscriptions5(self):
        test_retry_wrapper(self._test_get_shadow_failure_empty_subscriptions5)

    def _test_get_shadow_failure_empty_subscriptions311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _empty_subscription_topic_filters(options)))

    def test_get_shadow_failure_empty_subscriptions311(self):
        test_retry_wrapper(self._test_get_shadow_failure_empty_subscriptions311)

    def _test_get_shadow_failure_none_publish_topic5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _none_publish_topic(options)))

    def test_get_shadow_failure_none_publish_topic5(self):
        test_retry_wrapper(self._test_get_shadow_failure_none_publish_topic5)

    def _test_get_shadow_failure_none_publish_topic311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _none_publish_topic(options)))

    def test_get_shadow_failure_none_publish_topic311(self):
        test_retry_wrapper(self._test_get_shadow_failure_none_publish_topic311)

    def _test_get_shadow_failure_bad_publish_topic5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_get_shadow_future_failure_test(
            protocol_client, lambda options: _bad_publish_topic(options)))

    def test_get_shadow_failure_bad_publish_topic5(self):
        test_retry_wrapper(self._test_get_shadow_failure_bad_publish_topic5)

    def _test_get_shadow_failure_bad_publish_topic311(self):
        self._do_mqtt311_test(
            lambda protocol_client: self._do_get_shadow_future_failure_test(
                protocol_client,
                lambda options: _bad_publish_topic(options)))

    def test_get_shadow_failure_bad_publish_topic311(self):
        test_retry_wrapper(self._test_get_shadow_failure_bad_publish_topic311)

    def _test_get_shadow_failure_publish_topic_type_mismatch5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _type_mismatch_publish_topic(options)))

    def test_get_shadow_failure_publish_topic_type_mismatch5(self):
        test_retry_wrapper(self._test_get_shadow_failure_publish_topic_type_mismatch5)

    def _test_get_shadow_failure_publish_topic_type_mismatch311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _type_mismatch_publish_topic(options)))

    def test_get_shadow_failure_publish_topic_type_mismatch311(self):
        test_retry_wrapper(self._test_get_shadow_failure_publish_topic_type_mismatch311)

    def _test_get_shadow_failure_correlation_token_type_mismatch5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _type_mismatch_correlation_token(options)))

    def test_get_shadow_failure_correlation_token_type_mismatch5(self):
        test_retry_wrapper(self._test_get_shadow_failure_correlation_token_type_mismatch5)

    def _test_get_shadow_failure_correlation_token_type_mismatch311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_get_shadow_failure_test(
            protocol_client, lambda options: _type_mismatch_correlation_token(options)))

    def test_get_shadow_failure_correlation_token_type_mismatch311(self):
        test_retry_wrapper(self._test_get_shadow_failure_correlation_token_type_mismatch311)

    # ==============================================================
    #             streaming operation SUCCESS TEST CASES
    # ==============================================================

    def _test_streaming_operation_success5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_stream_success_test(protocol_client))

    def test_streaming_operation_success5(self):
        test_retry_wrapper(self._test_streaming_operation_success5)

    def _test_streaming_operation_success311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_stream_success_test(protocol_client))

    def test_streaming_operation_success311(self):
        test_retry_wrapper(self._test_streaming_operation_success311)

    # ==============================================================
    #             create_stream FAILURE TEST CASES
    # ==============================================================
    def _test_create_stream_failure_subscription_topic_filter_none5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_create_stream_failure_test(
            protocol_client, lambda options: _subscription_topic_filter_none(options)))

    def test_create_stream_failure_subscription_topic_filter_none5(self):
        test_retry_wrapper(self._test_create_stream_failure_subscription_topic_filter_none5)

    def _test_create_stream_failure_subscription_topic_filter_none311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_create_stream_failure_test(
            protocol_client, lambda options: _subscription_topic_filter_none(options)))

    def test_create_stream_failure_subscription_topic_filter_none311(self):
        test_retry_wrapper(self._test_create_stream_failure_subscription_topic_filter_none311)

    def _test_create_stream_failure_subscription_topic_filter_type_mismatch5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_create_stream_failure_test(
            protocol_client, lambda options: _type_mismatch_stream_subscription_topic_filter(options)))

    def test_create_stream_failure_subscription_topic_filter_type_mismatch5(self):
        test_retry_wrapper(self._test_create_stream_failure_subscription_topic_filter_type_mismatch5)

    def _test_create_stream_failure_subscription_topic_filter_type_mismatch311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_create_stream_failure_test(
            protocol_client, lambda options: _type_mismatch_stream_subscription_topic_filter(options)))

    def test_create_stream_failure_subscription_topic_filter_type_mismatch311(self):
        test_retry_wrapper(self._test_create_stream_failure_subscription_topic_filter_type_mismatch311)

    def _test_create_stream_failure_subscription_status_listener_type_mismatch5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_create_stream_failure_test(
            protocol_client, lambda options: _type_mismatch_subscription_status_listener(options)))

    def test_create_stream_failure_subscription_status_listener_type_mismatch5(self):
        test_retry_wrapper(self._test_create_stream_failure_subscription_status_listener_type_mismatch5)

    def _test_create_stream_failure_subscription_status_listener_type_mismatch311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_create_stream_failure_test(
            protocol_client, lambda options: _type_mismatch_subscription_status_listener(options)))

    def test_create_stream_failure_subscription_status_listener_type_mismatch311(self):
        test_retry_wrapper(self._test_create_stream_failure_subscription_status_listener_type_mismatch311)

    def _test_create_stream_failure_incoming_publish_listener_type_mismatch5(self):
        self._do_mqtt5_test(lambda protocol_client: self._do_create_stream_failure_test(
            protocol_client, lambda options: _type_mismatch_incoming_publish_listener(options)))

    def test_create_stream_failure_incoming_publish_listener_type_mismatch5(self):
        test_retry_wrapper(self._test_create_stream_failure_incoming_publish_listener_type_mismatch5)

    def _test_create_stream_failure_incoming_publish_listener_type_mismatch311(self):
        self._do_mqtt311_test(lambda protocol_client: self._do_create_stream_failure_test(
            protocol_client, lambda options: _type_mismatch_incoming_publish_listener(options)))

    def test_create_stream_failure_incoming_publish_listener_type_mismatch311(self):
        test_retry_wrapper(self._test_create_stream_failure_incoming_publish_listener_type_mismatch311)


if __name__ == 'main':
    unittest.main()
